from __future__ import annotations

from json import loads
from typing import Any, Dict, Iterable, Optional

from loguru import logger
from pydantic.v1.utils import deep_update
from pydash.objects import get
from sqlalchemy import Boolean, CheckConstraint, Column
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Session
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import Base, JSONTypeOverride
from fides.config import CONFIG, FidesConfig


class ApplicationConfig(Base):
    """
    Stores application config settings, set through different mechanisms, as JSON blobs in the DB.

    This is a single-row table. The single record describes global application config settings.
    """

    api_set = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=False,
        default={},
    )  # store as encrypted JSON blob since config settings may have sensitive data
    config_set = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=False,
        default={},
    )  # store as encrypted JSON blob since config settings may have sensitive data
    single_row = Column(
        Boolean, default=True, nullable=False
    )  # used to constrain table to one row

    CheckConstraint("single_row", name="single_row_check")

    @classmethod
    def create_or_update(  # type: ignore[override]
        cls,
        db: Session,
        *,
        data: Dict[str, Any],
        merge_updates: bool = True,
    ) -> ApplicationConfig:
        """
        Creates the config record if none exists, or updates the existing record.

        Here we effectively prevent more than a single record in the table.
        """
        existing_record = db.query(cls).first()
        if existing_record:
            updated_record = existing_record.update(
                db=db, data=data, merge_updates=merge_updates
            )
            return updated_record

        return cls.create(db=db, data=data)

    def update(self, db: Session, data: Dict[str, Any], merge_updates: bool = True) -> ApplicationConfig:  # type: ignore[override]
        """
        Updates the config record, merging contents of the particular JSON column that
        corresponds to the updated data.

        Given `data` dict should have either/both an `api_set` and/or `config_set` key at the top-level.
        Those keys will then point to `dict`s of data to update, i.e. merge, with the existing contents
        of the `api_set` or `config_set` columns, respectively.
        """
        if "api_set" in data:
            incoming_api_config = data["api_set"]
            if not isinstance(incoming_api_config, dict):
                raise ValueError("`api_set` column must be a dictionary")

            if merge_updates:
                # merge, the `api_set` dict
                self.api_set = deep_update(self.api_set, incoming_api_config)
            else:
                # replace the `api_set` dict
                self.api_set = incoming_api_config

        if "config_set" in data:
            incoming_config_config = data["config_set"]
            if not isinstance(incoming_config_config, dict):
                raise ValueError("`config_set` column must be a dictionary")

            if merge_updates:
                # update, i.e. merge, the `config_set` dict
                self.config_set = deep_update(self.config_set, incoming_config_config)
            else:
                self.config_set = incoming_config_config

        self.save(db=db)
        return self

    @classmethod
    def get_api_set(cls, db: Session) -> Dict[str, Any]:
        """
        Utility method to get the api_set config settings dict

        An empty `dict` will be returned if no config settings have been set through the API
        """
        config_record = db.query(cls).first()
        if config_record:
            return config_record.api_set
        return {}

    @classmethod
    def get_config_set(cls, db: Session) -> Dict[str, Any]:
        """
        Utility method to get the config_set config dict

        An empty `dict` will be returned if no config settings have been set through config
        """
        config_record = db.query(cls).first()
        if config_record:
            return config_record.config_set
        return {}

    @classmethod
    def update_api_set(
        cls,
        db: Session,
        api_set_dict: Dict[str, Any],
        merge_updates: bool = True,
    ) -> ApplicationConfig:
        """
        Utility method to set the `api_set` column on the `applicationconfig`
        db record with the provided dictionary of data.

        Depending on `merge_updates` param, updates either:
         - are *merged* with any existing `api_set` data in the db
         - replace any existing `api_set` data in the db
        """
        return cls.create_or_update(
            db, data={"api_set": api_set_dict}, merge_updates=merge_updates
        )

    @classmethod
    def clear_api_set(cls, db: Session) -> Optional[ApplicationConfig]:
        """
        Utility method to set the `api_set` column on the `applicationconfig`
        db record to an empty dict

        """
        existing_record = db.query(cls).first()
        if existing_record:
            existing_record.api_set = {}
            existing_record.save(db)
            return existing_record
        return None

    @classmethod
    def clear_config_set(cls, db: Session) -> Optional[ApplicationConfig]:
        """
        Utility method to set the `config_set` column on the `applicationconfig`
        db record to an empty dict

        """
        existing_record = db.query(cls).first()
        if existing_record:
            existing_record.config_set = {}
            existing_record.save(db)
            return existing_record
        return None

    @classmethod
    def update_config_set(cls, db: Session, config: FidesConfig) -> ApplicationConfig:
        """
        Utility method to set the `config_set` column on the `applicationconfig`
        db record by serializing the given `FidesConfig` to a JSON blob.
        """
        # need to do this deserialization round-trip to get us a dict
        # that's serializable as JSON in the db:
        # pydantic's `.dict()` is NOT serializable as JSON --
        # see https://github.com/pydantic/pydantic/issues/1409
        config_dict = loads(config.json())
        return cls.create_or_update(db, data={"config_set": config_dict})

    @classmethod
    def get_resolved_config_property(
        cls,
        db: Session,
        config_property: str,
        merge_values: bool = False,
        default_value: Any = None,
    ) -> Optional[Any]:
        """
        Gets the 'resolved' config property based on api-set and config-set configs.
        `config_property` is a dot-separated path to the config property,
        e.g. `notifications.notification_service_type`.

        Api-set values get priority over config-set, in case of conflict, unless `merge_values`
        is specified as `True`, in which case the proxy will attempt to _merge_ the api-set
        and config-set values. Note that only iterable config values (e.g. lists) can be merged!

        An error will be thrown if `merge_values` is used for a property with non-iterable values.
        """
        config_record = db.query(cls).first()
        if config_record:
            api_prop = get(config_record.api_set, config_property)
            config_prop = get(config_record.config_set, config_property, default_value)
            # if no api-set property found, fall back to config-set
            if api_prop is None:
                return config_prop

            # if we want to merge values and have a config property too
            if merge_values and config_prop is not None:
                # AND we have iterable api and config-set properties, then we try to merge
                if not isinstance(api_prop, Iterable) or not isinstance(
                    config_prop, Iterable
                ):
                    raise RuntimeError("Only iterable values can be merged!")
                return {value for value in (list(api_prop) + list(config_prop))}

            # otherwise, we just use the api-set property
            return api_prop

        logger.warning("No config record found!")
        return default_value



from __future__ import annotations

from collections import defaultdict
from enum import Enum
from functools import lru_cache
from os.path import dirname, join
from typing import Any, Dict, Iterable, List, Set, Union

import yaml
from pydantic import BaseModel
from sqlalchemy import (
    ARRAY,
    Boolean,
    CheckConstraint,
    Column,
    String,
    UniqueConstraint,
    insert,
    select,
    update,
)
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base, FidesBase
from fides.config.helpers import load_file

LOCATIONS_YAML_FILE_PATH = join(
    dirname(__file__),
    "../../data",
    "location",
    "locations.yml",
)


class LocationRegulationSelections(Base):
    """
    Persists application-wide location and regulation selections in the DB.

    This is a single-row table. The single record describes global, application-wide selections
    about enabled locations and regulations.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "location_regulation_selections"

    selected_locations = Column(
        ARRAY(String),
        nullable=False,
        default={},
    )
    selected_location_groups = (
        Column(  # Don't save this directly, it is saved as a side-effect of locations
            ARRAY(String),
            nullable=False,
            default={},
        )
    )
    selected_regulations = Column(
        ARRAY(String),
        nullable=False,
        default={},
    )
    single_row = Column(
        Boolean,
        default=True,
        nullable=False,
        unique=True,
    )  # used to constrain table to one row

    CheckConstraint("single_row", name="single_row_check")
    UniqueConstraint("single_row", name="single_row_unique")

    @classmethod
    def create_or_update(  # type: ignore[override]
        cls,
        db: Session,
        *,
        data: Dict[str, Any],
    ) -> FidesBase:
        """
        Creates the selections record if none exists, or updates the existing record.

        Here we effectively prevent more than a single record in the table.
        """
        existing_record = db.query(cls).first()
        if existing_record:
            updated_record = existing_record.update(
                db=db,
                data=data,
            )  # type: ignore[arg-type]
            return updated_record

        return cls.create(db=db, data=data)

    @classmethod
    async def create_or_update_async(
        cls,
        async_session: AsyncSession,
        *,
        data: Dict[str, Any],
    ) -> LocationRegulationSelections:
        """
        Creates the selections record if none exists, or updates the existing record.

        Here we effectively prevent more than a single record in the table.
        """
        async with async_session.begin():
            result = await async_session.execute(select(cls))  # type: ignore[arg-type]
            existing_record = result.scalars().first()
            if existing_record:
                await async_session.execute(
                    update(cls).where(cls.id == existing_record.id).values(data)  # type: ignore[arg-type]
                )
            else:
                await async_session.execute(insert(cls).values(data))  # type: ignore[arg-type]
            result = await async_session.execute(select(cls))  # type: ignore[arg-type]
            return result.scalars().first()

    @classmethod
    def set_selected_locations(
        cls,
        db: Session,
        selected_locations: Iterable[str],
    ) -> None:
        """Utility method to set the selected locations

        Calculates "location groups" from locations and saves this too. Since an allowed region on a Privacy Experience
        can be a location or a location group, precalculating location groups will make this a faster lookup.
        """
        selected_location_groups: Set[str] = group_locations_into_location_groups(
            selected_locations
        )

        cls.create_or_update(
            db,
            data={
                "selected_locations": set(selected_locations),
                "selected_location_groups": selected_location_groups,
            },
        )

    @classmethod
    async def set_selected_locations_async(
        cls,
        async_session: AsyncSession,
        selected_locations: Iterable[str],
    ) -> None:
        """Utility method to set the selected locations.  Location groups are calculated from locations and saved
        here as well"""
        selected_location_groups: Set[str] = group_locations_into_location_groups(
            selected_locations
        )

        await cls.create_or_update_async(
            async_session,
            data={
                "selected_locations": set(selected_locations),
                "selected_location_groups": selected_location_groups,
            },
        )

    @classmethod
    def get_selected_locations(
        cls,
        db: Session,
    ) -> Set[str]:
        """
        Utility method to get the selected_locations, returned as a Set.
        """
        record = db.query(cls).first()
        if record:
            return set(record.selected_locations)
        return set()

    @classmethod
    async def get_selected_locations_async(
        cls,
        async_session: AsyncSession,
    ) -> Set[str]:
        """
        Utility method to get the selected_locations, returned as a Set.
        """
        async with async_session.begin():
            results = await async_session.execute(select(cls))  # type: ignore[arg-type]
            record = results.scalars().first()
            if record:
                return set(record.selected_locations)
            return set()

    @classmethod
    def get_selected_location_groups(
        cls,
        db: Session,
    ) -> Set[str]:
        """
        Utility method to get the selected_locations_groups, returned as a Set.

        Location Groups aren't saved directly, but as a side-effect of saving locations
        """
        record = db.query(cls).first()
        if record:
            return set(record.selected_location_groups)
        return set()

    @classmethod
    async def get_selected_location_groups_async(
        cls,
        async_session: AsyncSession,
    ) -> Set[str]:
        """
        Utility method to get the selected_location_groups, returned as a Set.

        Location Groups aren't saved directly, but as a side-effect of saving locations
        """
        async with async_session.begin():
            results = await async_session.execute(select(cls))  # type: ignore[arg-type]
            record = results.scalars().first()
            if record:
                return set(record.selected_location_groups)
            return set()

    @classmethod
    def set_selected_regulations(
        cls,
        db: Session,
        selected_regulations: Iterable[str],
    ) -> None:
        """Utility method to set the selected regulations"""
        cls.create_or_update(
            db, data={"selected_regulations": set(selected_regulations)}
        )

    @classmethod
    async def set_selected_regulations_async(
        cls,
        async_session: AsyncSession,
        selected_regulations: Iterable[str],
    ) -> None:
        """Utility method to set the selected regulations"""
        await cls.create_or_update_async(
            async_session, data={"selected_regulations": set(selected_regulations)}
        )

    @classmethod
    def get_selected_regulations(
        cls,
        db: Session,
    ) -> Set[str]:
        """
        Utility method to get the selected_regulations, returned as a Set.
        """
        record = db.query(cls).first()
        if record:
            return set(record.selected_regulations)
        return set()

    @classmethod
    async def get_selected_regulations_async(
        cls,
        async_session: AsyncSession,
    ) -> Set[str]:
        """
        Utility method to get the selected_regulations, returned as a Set.
        """
        async with async_session.begin():
            results = await async_session.execute(select(cls))  # type: ignore[arg-type]
            record = results.scalars().first()
            if record:
                return set(record.selected_regulations)
            return set()


def group_locations_into_location_groups(saved_locations: Iterable[str]) -> Set[str]:
    """Combines saved Locations into Location groups if applicable
    All Locations must be present for a Location Group to be valid.
    """
    saved_location_groups: List[str] = []

    for location_group, mapped_locations in location_group_to_location.items():
        if mapped_locations.issubset(saved_locations):
            saved_location_groups.append(location_group)

    return set(saved_location_groups)


class Continent(Enum):
    """Enum of supported continents"""

    north_america = "North America"
    south_america = "South America"
    asia = "Asia"
    africa = "Africa"
    oceania = "Oceania"
    europe = "Europe"


class Selection(BaseModel):
    """Selection schema"""

    id: str
    selected: bool = False


class LocationRegulationBase(Selection):
    """Base Location Regulation Schema"""

    name: str
    continent: Continent
    default_selected: bool = False


class Location(LocationRegulationBase):
    """Location schema"""

    belongs_to: List[str] = []
    regulation: List[str] = []
    is_country: bool = False


class LocationGroup(Location):
    """
    Location Group schema, currently the same as a location
    """


@lru_cache(maxsize=1)
def load_locations() -> Dict[str, Location]:
    """Loads location dict based on yml file on disk"""
    with open(load_file([LOCATIONS_YAML_FILE_PATH]), "r", encoding="utf-8") as file:
        _locations = yaml.safe_load(file).get("locations", [])
        location_dict = {}
        for location in _locations:
            location_dict[location["id"]] = Location.model_validate(location)
        return location_dict


@lru_cache(maxsize=1)
def load_location_groups() -> Dict[str, LocationGroup]:
    """Loads location_groups dict based on yml file on disk"""
    with open(load_file([LOCATIONS_YAML_FILE_PATH]), "r", encoding="utf-8") as file:
        _location_groups = yaml.safe_load(file).get("location_groups", [])
        location_group_dict = {}
        for location_group in _location_groups:
            location_group_dict[location_group["id"]] = LocationGroup.model_validate(
                location_group
            )
        return location_group_dict


@lru_cache(maxsize=1)
def load_location_group_to_location() -> Dict[str, Set[str]]:
    """Returns a mapping of locations to location groups

    This is helpful when locations are saved, for use in mapping those locations to location groups.
    Our current API only supports locations being saved directly.
    """
    location_ids_mapped_to_location: Dict[str, Location] = load_locations()

    location_groups_to_location_set: Dict[str, Set[str]] = defaultdict(set)

    for location in location_ids_mapped_to_location.values():
        if not location.belongs_to:
            continue

        for belongs_to in location.belongs_to:
            location_groups_to_location_set[belongs_to].add(location.id)

    return location_groups_to_location_set


@lru_cache(maxsize=1)
def load_regulations() -> Dict[str, LocationRegulationBase]:
    """Loads regulations dict based on yml file on disk"""
    with open(load_file([LOCATIONS_YAML_FILE_PATH]), "r", encoding="utf-8") as file:
        _regulations = yaml.safe_load(file).get("regulations", [])
        regulation_dict = {}
        for regulation in _regulations:
            regulation_dict[regulation["id"]] = LocationRegulationBase.model_validate(
                regulation
            )
        return regulation_dict


locations_by_id: Dict[str, Location] = (
    load_locations()
)  # should only be accessed for read-only access
location_groups: Dict[str, LocationGroup] = (
    load_location_groups()
)  # should only be accessed for read-only access
location_group_to_location: Dict[str, Set[str]] = (
    load_location_group_to_location()
)  # should only be accessed for read-only access
default_selected_locations: Set[str] = {
    id for id, location in locations_by_id.items() if location.default_selected
}


def _load_privacy_notice_regions() -> Dict[str, Union[Location, LocationGroup]]:
    """Merge Location dictionary with LocationGroup Dictionary"""
    return {**locations_by_id, **location_groups}


privacy_notice_regions_by_id: Dict[str, Union[Location, LocationGroup]] = (
    _load_privacy_notice_regions()
)  # should only be accessed for read-only access


# dynamically create an enum based on definitions loaded from YAML
# This is a combination of "locations" and "location groups" for use on Privacy Experiences
PrivacyNoticeRegion = Enum(  # type: ignore[misc]
    "PrivacyNoticeRegion",
    [(location.id, location.id) for location in privacy_notice_regions_by_id.values()],
    type=str,
)

# Create a notice region enum that includes regions we no longer support but still preserve
# on historical records for auditing purposes
deprecated_gb_regions = ["gb_eng", "gb_sct", "gb_wls", "gb_nir"]
current_regions = {
    location.id: location.id for location in privacy_notice_regions_by_id.values()
}
current_regions.update({gb_region: gb_region for gb_region in deprecated_gb_regions})
DeprecatedNoticeRegion: Enum = Enum(  # type: ignore[misc]
    "DeprecatedNoticeRegion", current_regions
)


def filter_regions_by_location(
    db: Session,
    regions: List[PrivacyNoticeRegion],
) -> List[PrivacyNoticeRegion]:
    """Filter a list of PrivacyNoticeRegions to only ones that match at the configured Location or
    LocationGroup level.

    Only Experiences with these regions will be shown to the end-user!
    """

    saved_locations: Set[str] = LocationRegulationSelections.get_selected_locations(db)
    saved_location_groups: Set[str] = (
        LocationRegulationSelections.get_selected_location_groups(db)
    )
    multilevel_locations: Set[str] = saved_locations.union(saved_location_groups)

    # For backwards-compatibility, if no system-wide locations or location groups are set,
    # all PrivacyNoticeRegions are available for use on Experiences
    if not multilevel_locations:
        return regions  # type: ignore[attr-defined]

    return [region for region in regions if region.value in multilevel_locations]



from typing import Annotated, Any, Dict, List, Optional

import yaml
from fastapi import Depends, HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.params import Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.models import Dataset as FideslangDataset
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field
from pydantic import ValidationError as PydanticValidationError
from sqlalchemy import and_, not_, select
from sqlalchemy.orm import Session
from starlette.status import (
    HTTP_200_OK,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_415_UNSUPPORTED_MEDIA_TYPE,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from fides.api.api import deps
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.policy import Policy
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.dataset import (
    BulkPutDataset,
    DatasetConfigCtlDataset,
    DatasetConfigSchema,
    DatasetReachability,
    ValidateDatasetResponse,
)
from fides.api.schemas.privacy_request import TestPrivacyRequest
from fides.api.schemas.redis_cache import DatasetTestRequest
from fides.api.service.deps import get_dataset_config_service
from fides.api.util.api_router import APIRouter
from fides.common.api.scope_registry import (
    DATASET_CREATE_OR_UPDATE,
    DATASET_DELETE,
    DATASET_READ,
    DATASET_TEST,
)
from fides.common.api.v1.urn_registry import (
    CONNECTION_DATASETS,
    DATASET_BY_KEY,
    DATASET_CONFIG_BY_KEY,
    DATASET_CONFIGS,
    DATASET_INPUTS,
    DATASET_REACHABILITY,
    DATASET_VALIDATE,
    DATASETS,
    TEST_DATASET,
    V1_URL_PREFIX,
    YAML_DATASETS,
)
from fides.config import CONFIG
from fides.service.dataset.dataset_config_service import DatasetConfigService
from fides.service.dataset.dataset_service import DatasetNotFoundException

from fides.api.models.sql_models import (  # type: ignore[attr-defined] # isort: skip
    Dataset as CtlDataset,
)

X_YAML = "application/x-yaml"

router = APIRouter(tags=["Dataset Configs"], prefix=V1_URL_PREFIX)


# Helper method to inject the parent ConnectionConfig into these child routes
def _get_connection_config(
    connection_key: FidesKey, db: Session = Depends(deps.get_db)
) -> ConnectionConfig:
    logger.debug("Finding connection config with key '{}'", connection_key)
    connection_config = ConnectionConfig.get_by(db, field="key", value=connection_key)
    if not connection_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No connection config with key '{connection_key}'",
        )
    return connection_config


@router.put(
    DATASET_VALIDATE,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
    status_code=HTTP_200_OK,
    response_model=ValidateDatasetResponse,
)
def validate_dataset(
    dataset: FideslangDataset,
    dataset_config_service: DatasetConfigService = Depends(get_dataset_config_service),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> ValidateDatasetResponse:
    """
    Run validations against a dataset without attempting to save it to the database.

    Checks that:
    - all required fields are present, all field values are valid types
    - all DataCategory values reference known keys
    - etc.

    After validating, also tests to see if the dataset is traversable. Note that
    it's possible for a dataset to be valid but not traversable; this happens
    when a dataset is dependent on references to other datasets.

    Returns a 200 OK for all valid datasets, and a traversal_details object with
    information about the traversal (or traversal errors).
    """

    try:
        return dataset_config_service.validate_dataset_config(
            connection_config, dataset
        )
    except PydanticValidationError as e:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(e.errors(include_url=False, include_input=False)),
        )


@router.put(
    DATASET_CONFIGS,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_CREATE_OR_UPDATE])],
    status_code=HTTP_200_OK,
    response_model=BulkPutDataset,
)
def put_dataset_configs(
    dataset_pairs: Annotated[List[DatasetConfigCtlDataset], Field(max_length=50)],  # type: ignore
    db: Session = Depends(deps.get_db),
    dataset_config_service: DatasetConfigService = Depends(get_dataset_config_service),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> BulkPutDataset:
    """
    Endpoint to create, update, or remove DatasetConfigs by passing in pairs of:
    1) A DatasetConfig fides_key
    2) The corresponding CtlDataset fides_key which stores the bulk of the actual dataset

    The CtlDataset contents are retrieved for extra validation before linking this
    to the DatasetConfig.

    Note: Any existing DatasetConfigs not specified in the dataset pairs will be deleted.
    """

    # first delete any dataset configs not in the dataset pairs
    existing_config_keys = set(
        db.execute(
            select([DatasetConfig.fides_key]).where(
                DatasetConfig.connection_config_id == connection_config.id
            )
        )
        .scalars()
        .all()
    )

    requested_config_keys = {pair.fides_key for pair in dataset_pairs}
    config_keys_to_remove = existing_config_keys - requested_config_keys

    if config_keys_to_remove:
        db.query(DatasetConfig).filter(
            DatasetConfig.connection_config_id == connection_config.id,
            DatasetConfig.fides_key.in_(config_keys_to_remove),
        ).delete(synchronize_session=False)
        db.commit()

    # reuse the existing patch logic once we've removed the unused dataset configs
    return patch_dataset_configs(
        dataset_pairs, dataset_config_service, connection_config
    )


@router.patch(
    DATASET_CONFIGS,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_CREATE_OR_UPDATE])],
    status_code=HTTP_200_OK,
    response_model=BulkPutDataset,
)
def patch_dataset_configs(
    dataset_pairs: Annotated[List[DatasetConfigCtlDataset], Field(max_length=50)],  # type: ignore
    dataset_config_service: DatasetConfigService = Depends(get_dataset_config_service),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> BulkPutDataset:
    """
    Endpoint to create or update DatasetConfigs by passing in pairs of:
    1) A DatasetConfig fides_key
    2) The corresponding CtlDataset fides_key which stores the bulk of the actual dataset

    The CtlDataset contents are retrieved for extra validation before linking this
    to the DatasetConfig.
    """
    try:
        return dataset_config_service.bulk_create_or_update_dataset_configs(
            connection_config,
            dataset_pairs,
        )
    except DatasetNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=str(e),
        )
    except PydanticValidationError as e:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(e.errors(include_url=False, include_input=False)),
        )


@router.patch(
    CONNECTION_DATASETS,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_CREATE_OR_UPDATE])],
    status_code=HTTP_200_OK,
    response_model=BulkPutDataset,
)
def patch_datasets(
    datasets: Annotated[List[FideslangDataset], Field(max_length=50)],  # type: ignore
    dataset_config_service: DatasetConfigService = Depends(get_dataset_config_service),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> BulkPutDataset:
    """
    Given a list of dataset elements, create or update corresponding Dataset objects
    or report failure

    This endpoint upserts the DatasetConfig and associated CTL Dataset.  Will shortly be deprecated.

    If the fides_key for a given DatasetConfig exists, it will be treated as an update.
    Otherwise, a new DatasetConfig will be created.
    """

    try:
        return dataset_config_service.bulk_create_or_update_dataset_configs(
            connection_config, datasets
        )
    except PydanticValidationError as e:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(e.errors(include_url=False, include_input=False)),
        )


@router.patch(
    YAML_DATASETS,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_CREATE_OR_UPDATE])],
    status_code=HTTP_200_OK,
    response_model=BulkPutDataset,
)
async def patch_yaml_datasets(
    request: Request,
    dataset_config_service: DatasetConfigService = Depends(get_dataset_config_service),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> BulkPutDataset:
    """
    Bulk create or update datasets from YAML format.
    Accepts application/x-yaml content type.
    """
    if request.headers.get("content-type") != X_YAML:
        raise HTTPException(
            status_code=HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            detail=f"Content type must be {X_YAML}",
        )

    try:
        # Read and parse YAML content
        body = await request.body()
        yaml_request_body = yaml.safe_load(body)

        # Convert YAML datasets to Dataset objects
        datasets = [
            FideslangDataset.model_validate(dataset_dict)
            for dataset_dict in yaml_request_body["dataset"]
        ]

        return dataset_config_service.bulk_create_or_update_dataset_configs(
            connection_config, datasets
        )

    except yaml.MarkedYAMLError as e:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Invalid YAML format: {str(e)}",
        )
    except PydanticValidationError as e:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Dataset validation failed: {str(e)}",
        )


@router.get(
    CONNECTION_DATASETS,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
    response_model=Page[FideslangDataset],
    deprecated=True,
)
def get_datasets(
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> AbstractPage[FideslangDataset]:
    """Returns all CTL datasets attached to the ConnectionConfig via the Dataset Config.

    Soon to be deprecated.
    """

    logger.debug(
        "Finding all datasets for connection '{}' with pagination params {}",
        connection_config.key,
        params,
    )
    dataset_configs = DatasetConfig.filter(
        db=db, conditions=(DatasetConfig.connection_config_id == connection_config.id)
    ).order_by(DatasetConfig.created_at.desc())

    # Generate the paginated results, but don't return them as-is. Instead,
    # modify the items array to be just the Dataset instead of the full
    # DatasetConfig. This has to be done *afterwards* to ensure that the
    # paginated query is handled by paginate()
    paginated_results = paginate(dataset_configs, params=params)
    paginated_results.items = [  # type: ignore
        dataset_config.ctl_dataset for dataset_config in paginated_results.items  # type: ignore
    ]
    return paginated_results


@router.get(
    DATASET_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
    response_model=FideslangDataset,
    deprecated=True,
)
def get_dataset(
    dataset_key: FidesKey,
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> FideslangDataset:
    """Returns a single ctl dataset linked to the given DatasetConfig.

    Soon to be deprecated
    """

    logger.debug(
        "Finding dataset '{}' for connection '{}'", dataset_key, connection_config.key
    )
    dataset_config = DatasetConfig.filter(
        db=db,
        conditions=(
            (DatasetConfig.connection_config_id == connection_config.id)
            & (DatasetConfig.fides_key == dataset_key)
        ),
    ).first()
    if not dataset_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset with fides_key '{dataset_key}' and connection key {connection_config.key}'",
        )
    return dataset_config.ctl_dataset


@router.get(
    DATASET_CONFIGS,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
    response_model=Page[DatasetConfigSchema],
)
def get_dataset_configs(
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> AbstractPage[DatasetConfig]:
    """Returns all Dataset Configs attached to current Connection Config."""

    logger.debug(
        "Finding all dataset configs for connection '{}' with pagination params {}",
        connection_config.key,
        params,
    )
    dataset_configs = DatasetConfig.filter(
        db=db, conditions=(DatasetConfig.connection_config_id == connection_config.id)
    ).order_by(DatasetConfig.created_at.desc())

    return paginate(dataset_configs, params)


@router.get(
    DATASET_CONFIG_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
    response_model=DatasetConfigSchema,
)
def get_dataset_config(
    dataset_key: FidesKey,
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> DatasetConfig:
    """Returns the specific Dataset Config linked to the Connection Config."""

    logger.debug(
        "Finding dataset config '{}' for connection '{}'",
        dataset_key,
        connection_config.key,
    )
    dataset_config = DatasetConfig.filter(
        db=db,
        conditions=(
            (DatasetConfig.connection_config_id == connection_config.id)
            & (DatasetConfig.fides_key == dataset_key)
        ),
    ).first()
    if not dataset_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset config with fides_key '{dataset_key}' and connection key {connection_config.key}'",
        )
    return dataset_config


@router.delete(
    DATASET_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_DELETE])],
    status_code=HTTP_204_NO_CONTENT,
)
def delete_dataset(
    dataset_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
) -> None:
    """Removes the DatasetConfig based on the given key."""

    logger.info(
        "Finding dataset '{}' for connection '{}'", dataset_key, connection_config.key
    )
    dataset_config = DatasetConfig.filter(
        db=db,
        conditions=(
            (DatasetConfig.connection_config_id == connection_config.id)
            & (DatasetConfig.fides_key == dataset_key)
        ),
    ).first()
    if not dataset_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset with fides_key '{dataset_key}' and connection_key '{connection_config.key}'",
        )

    logger.info(
        "Deleting dataset '{}' for connection '{}'", dataset_key, connection_config.key
    )
    dataset_config.delete(db)


@router.get(
    f"/filter{DATASETS}",
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
    response_model=List[FideslangDataset],
    deprecated=True,
)
def get_ctl_datasets(
    db: Session = Depends(deps.get_db),
    remove_saas_datasets: bool = True,
    only_unlinked_datasets: bool = False,
) -> List[FideslangDataset]:
    """
    Deprecated. Use `GET /datasets` instead.
    Returns all CTL datasets .
    """

    logger.debug(
        f"Finding all datasets {remove_saas_datasets=} {only_unlinked_datasets=}"
    )
    filters = []
    if only_unlinked_datasets:
        unlinked_subquery = select([DatasetConfig.ctl_dataset_id])
        filters.append(not_(CtlDataset.id.in_(unlinked_subquery)))

    if remove_saas_datasets:
        saas_subquery = (
            select([ConnectionConfig.saas_config["fides_key"].astext])
            .select_from(ConnectionConfig)  # type: ignore[arg-type]
            .where(ConnectionConfig.saas_config.is_not(None))  # type: ignore[attr-defined]
        )
        filters.append(not_(CtlDataset.fides_key.in_(saas_subquery)))

    conditions = []

    if len(filters) > 0:
        if len(filters) == 1:
            conditions.append(filters[0])
        else:
            conditions.append(and_(*filters))

    query = db.query(CtlDataset).filter(*conditions).order_by(CtlDataset.name.desc())
    datasets = query.all()

    return datasets


@router.get(
    DATASET_INPUTS,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
)
def dataset_identities_and_references(
    *,
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
    dataset_key: FidesKey,
) -> Dict[str, Any]:
    """
    Returns a dictionary containing the immediate identity and dataset reference
    dependencies for the given dataset.
    """

    dataset_config = DatasetConfig.filter(
        db=db,
        conditions=(
            (DatasetConfig.connection_config_id == connection_config.id)
            & (DatasetConfig.fides_key == dataset_key)
        ),
    ).first()
    if not dataset_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset config with fides_key '{dataset_key}'",
        )

    inputs = dataset_config.get_identities_and_references()
    return {input: None for input in inputs}


@router.get(
    DATASET_REACHABILITY,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_READ])],
    response_model=DatasetReachability,
)
def dataset_reachability(
    *,
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
    dataset_config_service: DatasetConfigService = Depends(get_dataset_config_service),
    dataset_key: FidesKey,
    policy_key: Optional[FidesKey] = None,
) -> Dict[str, Any]:
    """
    Returns a dictionary containing the immediate identity and dataset reference
    dependencies for the given dataset.
    """

    dataset_config = DatasetConfig.filter(
        db=db,
        conditions=(
            (DatasetConfig.connection_config_id == connection_config.id)
            & (DatasetConfig.fides_key == dataset_key)
        ),
    ).first()
    if not dataset_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset config with fides_key '{dataset_key}'",
        )

    access_policy = (
        Policy.get_by(db, field="key", value=policy_key) if policy_key else None
    )
    if policy_key and not access_policy:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f'Policy with key "{policy_key}" not found',
        )

    reachable, details = dataset_config_service.get_dataset_reachability(
        dataset_config, access_policy
    )
    return {"reachable": reachable, "details": details}


@router.post(
    TEST_DATASET,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[DATASET_TEST])],
    response_model=TestPrivacyRequest,
)
def test_connection_datasets(
    *,
    db: Session = Depends(deps.get_db),
    dataset_config_service: DatasetConfigService = Depends(get_dataset_config_service),
    connection_config: ConnectionConfig = Depends(_get_connection_config),
    dataset_key: FidesKey,
    test_request: DatasetTestRequest,
) -> Dict[str, Any]:

    if not CONFIG.security.dsr_testing_tools_enabled:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="DSR testing tools are not enabled.",
        )

    dataset_config = DatasetConfig.filter(
        db=db,
        conditions=(
            (DatasetConfig.connection_config_id == connection_config.id)
            & (DatasetConfig.fides_key == dataset_key)
        ),
    ).first()
    if not dataset_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset config with fides_key '{dataset_key}'",
        )

    access_policy = Policy.get_by(db, field="key", value=test_request.policy_key)
    if not access_policy:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f'Policy with key "{test_request.policy_key}" not found',
        )

    privacy_request = dataset_config_service.run_test_access_request(
        access_policy,
        dataset_config,
        input_data=test_request.identities.data,
    )
    return {"privacy_request_id": privacy_request.id}



from contextlib import contextmanager
from typing import Generator

from fastapi import Depends
from sqlalchemy.orm import Session

from fides.api.common_exceptions import RedisNotConfigured
from fides.api.db.session import get_db_engine, get_db_session
from fides.api.util.cache import get_cache as get_redis_connection
from fides.config import CONFIG, FidesConfig
from fides.config import get_config as get_app_config
from fides.config.config_proxy import ConfigProxy

_engine = None


def get_config() -> FidesConfig:
    """Returns the config for use in dependency injection."""
    return get_app_config()


def get_db() -> Generator:
    """Return our database session"""
    try:
        db = get_api_session()
        yield db
    finally:
        db.close()


@contextmanager
def get_autoclose_db_session() -> Generator[Session, None, None]:
    """
    Return a database session as a context manager that automatically closes when the context exits.

    Unlike get_api_session which is managed by FastAPI's dependency injection,
    this context manager explicitly closes the session when exiting the context.
    Use this when you need manual control over the session lifecycle outside of API endpoints.
    """
    try:
        db = get_api_session()
        yield db
    finally:
        db.close()


def get_api_session() -> Session:
    """Gets the shared database session to use for API functionality"""
    global _engine  # pylint: disable=W0603
    if not _engine:
        _engine = get_db_engine(
            config=CONFIG,
            pool_size=CONFIG.database.api_engine_pool_size,
            max_overflow=CONFIG.database.api_engine_max_overflow,
            keepalives_idle=CONFIG.database.api_engine_keepalives_idle,
            keepalives_interval=CONFIG.database.api_engine_keepalives_interval,
            keepalives_count=CONFIG.database.api_engine_keepalives_count,
        )
    SessionLocal = get_db_session(CONFIG, engine=_engine)
    db = SessionLocal()
    return db


def get_config_proxy(db: Session = Depends(get_db)) -> ConfigProxy:
    return ConfigProxy(db)


def get_cache() -> Generator:
    """Return a connection to our Redis cache"""
    if not CONFIG.redis.enabled:
        raise RedisNotConfigured(
            "Application redis cache required, but it is currently disabled! Please update your application configuration to enable integration with a Redis cache."
        )
    yield get_redis_connection()



from typing import ClassVar, List

from pydantic import Field

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class HttpsSchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed to connect to a client api"""

    url: str
    authorization: str = Field(json_schema_extra={"sensitive": True})

    _required_components: ClassVar[List[str]] = ["url", "authorization"]


class HttpsDocsSchema(HttpsSchema, NoValidationSchema):
    """HTTPS Secrets Schema for API Docs"""



"""Module for interaction with User endpoints/commands."""

import json
from typing import Dict, List, Tuple

import requests
from fideslang.validation import FidesKey
from pydantic import EmailStr

from fides.api.cryptography.cryptographic_util import str_to_b64_str
from fides.common.utils import echo_green, echo_red, handle_cli_response
from fides.config import CONFIG
from fides.core.utils import (
    Credentials,
    get_auth_header,
    get_credentials_path,
    read_credentials_file,
    write_credentials_file,
)

CREATE_USER_PATH = "/api/v1/user"
LOGIN_PATH = "/api/v1/login"
USER_PERMISSIONS_PATH = "/api/v1/user/{}/permission"
SYSTEM_MANAGER_PATH = "/api/v1/user/{}/system-manager"


def get_access_token(username: str, password: str, server_url: str) -> Tuple[str, str]:
    """
    Get a user access token from the webserver.
    """
    payload = {
        "username": username,
        "password": str_to_b64_str(password),
    }

    response = requests.post(server_url + LOGIN_PATH, json=payload)
    handle_cli_response(response, verbose=False)
    user_id: str = response.json()["user_data"]["id"]
    access_token: str = response.json()["token_data"]["access_token"]
    return (user_id, access_token)


def create_user(
    username: str,
    password: str,
    email_address: EmailStr,
    first_name: str,
    last_name: str,
    auth_header: Dict[str, str],
    server_url: str,
) -> requests.Response:
    """Create a user."""
    request_data = {
        "username": username,
        "password": str_to_b64_str(password),
        "email_address": email_address,
        "first_name": first_name,
        "last_name": last_name,
    }
    response = requests.post(
        server_url + CREATE_USER_PATH,
        headers=auth_header,
        data=json.dumps(request_data),
    )
    handle_cli_response(response, verbose=False)
    return response


def get_user_permissions(
    user_id: str, auth_header: Dict[str, str], server_url: str
) -> Tuple[List[str], List[str]]:
    """
    Return a tuple of the total scopes the user has inherited via their roles, plus their roles
    """
    get_permissions_path = USER_PERMISSIONS_PATH.format(user_id)
    response = requests.get(
        server_url + get_permissions_path,
        headers=auth_header,
    )

    handle_cli_response(response, verbose=False)

    return (
        response.json()["total_scopes"],
        response.json()["roles"],
    )


def get_systems_managed_by_user(
    user_id: str, auth_header: Dict[str, str], server_url: str
) -> List[FidesKey]:
    """
    List all of the systems for which the current user is directly assigned
    """
    get_systems_path = SYSTEM_MANAGER_PATH.format(user_id)
    response = requests.get(
        server_url + get_systems_path,
        headers=auth_header,
    )

    handle_cli_response(response, verbose=False)
    return [system["fides_key"] for system in response.json()]


def update_user_permissions(
    user_id: str,
    auth_header: Dict[str, str],
    server_url: str,
    roles: List[str],
) -> requests.Response:
    """
    Update user permissions for a given user.
    """
    request_data = {"id": user_id, "roles": roles}
    set_permissions_path = USER_PERMISSIONS_PATH.format(user_id)
    response = requests.put(
        server_url + set_permissions_path,
        headers=auth_header,
        json=request_data,
    )
    handle_cli_response(response=response, verbose=False)
    return response


def create_command(
    username: str,
    password: str,
    email_address: EmailStr,
    first_name: str,
    last_name: str,
    server_url: str,
) -> None:
    """
    Given new user information, create a new user via the API using
    the local credentials file.
    """

    auth_header = get_auth_header()
    user_response = create_user(
        username=username,
        password=password,
        email_address=email_address,
        first_name=first_name,
        last_name=last_name,
        auth_header=auth_header,
        server_url=server_url,
    )
    user_id = user_response.json()["id"]
    new_user_roles = CONFIG.security.root_user_roles
    update_user_permissions(
        user_id=user_id,
        auth_header=auth_header,
        server_url=server_url,
        roles=new_user_roles,
    )
    echo_green(f"User: '{username}' created and assigned permissions: {new_user_roles}")


def login_command(username: str, password: str, server_url: str) -> str:
    """
    Given a username and password, request an access_token from the API and
    store all user information in a local credentials file.
    """
    user_id, access_token = get_access_token(
        username=username, password=password, server_url=server_url
    )
    echo_green(f"Logged in as user: {username}")
    credentials = Credentials(
        username=username, user_id=user_id, access_token=access_token
    )
    credentials_path = get_credentials_path()
    write_credentials_file(credentials, credentials_path)
    echo_green(f"Credentials file written to: {credentials_path}")
    return credentials_path


def get_permissions_command(server_url: str) -> None:
    """
    Get user permissions from the API.
    """
    credentials_path = get_credentials_path()
    try:
        credentials = read_credentials_file(credentials_path)
    except FileNotFoundError:
        echo_red(f"No credentials file found at path: {credentials_path}")
        raise SystemExit(1)

    user_id = credentials.user_id
    auth_header = get_auth_header()
    total_scopes, roles = get_user_permissions(user_id, auth_header, server_url)
    systems: List[FidesKey] = get_systems_managed_by_user(
        user_id, auth_header, server_url
    )

    print("Roles:")
    for role in roles:
        print(f"\t{role}")

    print("Associated scopes:")
    for scope in total_scopes:
        print(f"\t{scope}")

    print("Systems Under Management:")
    for system in systems:
        print(f"\t{system}")



"""Add privacy preference table

Revision ID: 8188ddbf8cae
Revises: ff782b0dc07e
Create Date: 2023-04-11 17:03:07.605408

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "8188ddbf8cae"
down_revision = "ff782b0dc07e"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "privacypreferencehistory",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "affected_system_status",
            postgresql.JSONB(astext_type=sa.Text()),
            server_default="{}",
            nullable=True,
        ),
        sa.Column(
            "email",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("hashed_email", sa.String(), nullable=True),
        sa.Column("hashed_phone_number", sa.String(), nullable=True),
        sa.Column(
            "phone_number",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "preference",
            sa.Enum("opt_in", "opt_out", "acknowledge", name="userconsentpreference"),
            nullable=False,
        ),
        sa.Column("privacy_notice_history_id", sa.String(), nullable=False),
        sa.Column("privacy_request_id", sa.String(), nullable=True),
        sa.Column("provided_identity_id", sa.String(), nullable=True),
        sa.Column("relevant_systems", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("request_origin", sa.String(), nullable=True),
        sa.Column(
            "secondary_user_ids",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("url_recorded", sa.String(), nullable=True),
        sa.Column(
            "user_agent",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("user_geography", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["privacy_notice_history_id"],
            ["privacynoticehistory.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_request_id"], ["privacyrequest.id"], ondelete="SET NULL"
        ),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_hashed_email"),
        "privacypreferencehistory",
        ["hashed_email"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_hashed_phone_number"),
        "privacypreferencehistory",
        ["hashed_phone_number"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_id"),
        "privacypreferencehistory",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_preference"),
        "privacypreferencehistory",
        ["preference"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_privacy_notice_history_id"),
        "privacypreferencehistory",
        ["privacy_notice_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_privacy_request_id"),
        "privacypreferencehistory",
        ["privacy_request_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_provided_identity_id"),
        "privacypreferencehistory",
        ["provided_identity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_user_geography"),
        "privacypreferencehistory",
        ["user_geography"],
        unique=False,
    )
    op.create_table(
        "currentprivacypreference",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "preference",
            sa.Enum("opt_in", "opt_out", "acknowledge", name="userconsentpreference"),
            nullable=False,
        ),
        sa.Column("provided_identity_id", sa.String(), nullable=True),
        sa.Column("privacy_notice_id", sa.String(), nullable=False),
        sa.Column("privacy_notice_history_id", sa.String(), nullable=False),
        sa.Column("privacy_preference_history_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["privacy_notice_history_id"],
            ["privacynoticehistory.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_id"],
            ["privacynotice.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_preference_history_id"],
            ["privacypreferencehistory.id"],
        ),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "provided_identity_id", "privacy_notice_id", name="identity_privacy_notice"
        ),
    )
    op.create_index(
        op.f("ix_currentprivacypreference_id"),
        "currentprivacypreference",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_preference"),
        "currentprivacypreference",
        ["preference"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_privacy_notice_history_id"),
        "currentprivacypreference",
        ["privacy_notice_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_privacy_notice_id"),
        "currentprivacypreference",
        ["privacy_notice_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_privacy_preference_history_id"),
        "currentprivacypreference",
        ["privacy_preference_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_provided_identity_id"),
        "currentprivacypreference",
        ["provided_identity_id"],
        unique=False,
    )
    op.add_column(
        "privacynotice", sa.Column("internal_description", sa.String(), nullable=True)
    )
    op.add_column(
        "privacynotice",
        sa.Column(
            "displayed_in_overlay", sa.Boolean(), nullable=False, server_default="f"
        ),
    )
    op.add_column(
        "privacynotice",
        sa.Column("displayed_in_api", sa.Boolean(), nullable=False, server_default="f"),
    )
    op.drop_column("privacynotice", "displayed_in_privacy_modal")
    op.drop_column("privacynotice", "displayed_in_banner")
    op.add_column(
        "privacynoticehistory",
        sa.Column("internal_description", sa.String(), nullable=True),
    )
    op.add_column(
        "privacynoticehistory",
        sa.Column(
            "displayed_in_overlay", sa.Boolean(), nullable=False, server_default="f"
        ),
    )
    op.add_column(
        "privacynoticehistory",
        sa.Column("displayed_in_api", sa.Boolean(), nullable=False, server_default="f"),
    )
    op.drop_column("privacynoticehistory", "displayed_in_privacy_modal")
    op.drop_column("privacynoticehistory", "displayed_in_banner")

    # Reverting server defaults that were added just for adding new fields
    op.alter_column("privacynoticehistory", "displayed_in_overlay", server_default=None)
    op.alter_column("privacynoticehistory", "displayed_in_api", server_default=None)
    op.alter_column("privacynotice", "displayed_in_overlay", server_default=None)
    op.alter_column("privacynotice", "displayed_in_api", server_default=None)


def downgrade():
    op.add_column(
        "privacynoticehistory",
        sa.Column(
            "displayed_in_banner", sa.BOOLEAN(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacynoticehistory",
        sa.Column(
            "displayed_in_privacy_modal",
            sa.BOOLEAN(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_column("privacynoticehistory", "displayed_in_api")
    op.drop_column("privacynoticehistory", "displayed_in_overlay")
    op.drop_column("privacynoticehistory", "internal_description")
    op.add_column(
        "privacynotice",
        sa.Column(
            "displayed_in_banner", sa.BOOLEAN(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacynotice",
        sa.Column(
            "displayed_in_privacy_modal",
            sa.BOOLEAN(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_column("privacynotice", "displayed_in_api")
    op.drop_column("privacynotice", "displayed_in_overlay")
    op.drop_column("privacynotice", "internal_description")
    op.drop_index(
        op.f("ix_currentprivacypreference_provided_identity_id"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_privacy_preference_history_id"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_privacy_notice_id"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_privacy_notice_history_id"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_preference"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_id"), table_name="currentprivacypreference"
    )
    op.drop_table("currentprivacypreference")
    op.drop_index(
        op.f("ix_privacypreferencehistory_user_geography"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_provided_identity_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_privacy_request_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_privacy_notice_history_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_preference"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_id"), table_name="privacypreferencehistory"
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_hashed_phone_number"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_hashed_email"),
        table_name="privacypreferencehistory",
    )
    op.drop_table("privacypreferencehistory")
    op.execute("drop type userconsentpreference")



"""adding custom privacy request field table

Revision ID: ad2e3f9a6850
Revises: 1ea164cee8bc
Create Date: 2023-09-06 20:09:10.188690

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "ad2e3f9a6850"
down_revision = "1ea164cee8bc"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "custom_privacy_request_field",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("privacy_request_id", sa.String(), nullable=True),
        sa.Column("field_name", sa.String(), nullable=False),
        sa.Column("field_label", sa.String(), nullable=False),
        sa.Column("hashed_value", sa.String(), nullable=True),
        sa.Column(
            "encrypted_value",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["privacy_request_id"],
            ["privacyrequest.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_custom_privacy_request_field_hashed_value"),
        "custom_privacy_request_field",
        ["hashed_value"],
        unique=False,
    )
    op.create_index(
        op.f("ix_custom_privacy_request_field_id"),
        "custom_privacy_request_field",
        ["id"],
        unique=False,
    )
    op.add_column(
        "privacyrequest",
        sa.Column(
            "custom_privacy_request_fields_approved_by", sa.String(), nullable=True
        ),
    )
    op.add_column(
        "privacyrequest",
        sa.Column(
            "custom_privacy_request_fields_approved_at",
            sa.DateTime(timezone=True),
            nullable=True,
        ),
    )
    op.create_foreign_key(
        "privacyrequest_custom_privacy_request_fields_approved_by_fkey",
        "privacyrequest",
        "fidesuser",
        ["custom_privacy_request_fields_approved_by"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    op.drop_constraint(
        "privacyrequest_custom_privacy_request_fields_approved_by_fkey",
        "privacyrequest",
        type_="foreignkey",
    )
    op.drop_column("privacyrequest", "custom_privacy_request_fields_approved_at")
    op.drop_column("privacyrequest", "custom_privacy_request_fields_approved_by")
    op.drop_index(
        op.f("ix_custom_privacy_request_field_id"),
        table_name="custom_privacy_request_field",
    )
    op.drop_index(
        op.f("ix_custom_privacy_request_field_hashed_value"),
        table_name="custom_privacy_request_field",
    )
    op.drop_table("custom_privacy_request_field")



"""
These utils are designed to be safe to use across Fides, with no potential for circular dependencies.

These utils should only import from 3rd-party libraries, with zero imports
from local Fides modules.
"""

import json
import pprint
import sys
from functools import partial
from json.decoder import JSONDecodeError
from typing import Dict, Union

import click
import requests
from loguru import logger

echo_red = partial(click.secho, fg="red", bold=True)
echo_green = partial(click.secho, fg="green", bold=True)


def print_divider(character: str = "-", character_length: int = 10) -> None:
    """
    Returns a consistent visual/textual divider to make terminal
    output more human-readable.
    """
    print(character * character_length)


def pretty_echo(dict_object: Union[Dict, str], color: str = "white") -> None:
    """
    Given a dict-like object and a color, pretty click echo it.
    """
    click.secho(pprint.pformat(dict_object, indent=2, width=80, compact=True), fg=color)


def handle_cli_response(
    response: requests.Response, verbose: bool = True
) -> requests.Response:
    """Viewable CLI response"""
    if response.status_code >= 200 and response.status_code <= 299:
        if verbose:
            pretty_echo(response.json(), "green")
    else:
        try:
            pretty_echo(response.json(), "red")
        except json.JSONDecodeError:
            click.secho(response.text, fg="red")
        finally:
            sys.exit(1)
    return response


def check_response_auth(response: requests.Response) -> requests.Response:
    """
    Verify that a response object is 'ok', otherwise print the error and raise
    an exception.
    """
    if response.status_code in [401, 403]:
        echo_red("Authorization Error: please try 'fides user login' and try again.")
        raise SystemExit(1)
    return response


def check_response(response: requests.Response) -> requests.Response:
    """
    Check that a response has valid JSON.
    """

    try:
        response.json()
    except JSONDecodeError as json_error:
        logger.error(response.status_code)
        logger.error(response.text)
        raise json_error

    return response



# pylint: disable=too-many-branches,too-many-lines, too-many-statements

import csv
import io
import json
from collections import defaultdict
from datetime import datetime
from typing import (
    Annotated,
    Any,
    DefaultDict,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Union,
)

import sqlalchemy
from fastapi import Body, Depends, HTTPException, Security
from fastapi.encoders import jsonable_encoder
from fastapi.params import Query as FastAPIQuery
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from loguru import logger
from pydantic import Field
from pydantic import ValidationError as PydanticValidationError
from sqlalchemy import cast, column, null, or_, select
from sqlalchemy.orm import Query, Session
from sqlalchemy.sql.expression import nullslast
from starlette.responses import StreamingResponse
from starlette.status import (
    HTTP_200_OK,
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from fides.api.api import deps
from fides.api.api.v1.endpoints.dataset_config_endpoints import _get_connection_config
from fides.api.api.v1.endpoints.manual_webhook_endpoints import (
    get_access_manual_webhook_or_404,
)
from fides.api.common_exceptions import (
    FidesopsException,
    IdentityVerificationException,
    ManualWebhookFieldsUnset,
    NoCachedManualWebhookEntry,
    TraversalError,
    ValidationError,
)
from fides.api.graph.config import CollectionAddress
from fides.api.graph.graph import DatasetGraph
from fides.api.graph.traversal import Traversal
from fides.api.models.audit_log import AuditLog, AuditLogAction
from fides.api.models.client import ClientDetail
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.manual_webhook import AccessManualWebhook
from fides.api.models.policy import Policy, PolicyPreWebhook, Rule
from fides.api.models.pre_approval_webhook import (
    PreApprovalWebhook,
    PreApprovalWebhookReply,
)
from fides.api.models.privacy_request import (
    EXITED_EXECUTION_LOG_STATUSES,
    CustomPrivacyRequestField,
    ExecutionLog,
    PrivacyRequest,
    PrivacyRequestNotifications,
    ProvidedIdentity,
    RequestTask,
)
from fides.api.oauth.utils import (
    verify_callback_oauth_policy_pre_webhook,
    verify_callback_oauth_pre_approval_webhook,
    verify_oauth_client,
    verify_request_task_callback,
)
from fides.api.schemas.dataset import CollectionAddressResponse, DryRunDatasetResponse
from fides.api.schemas.external_https import PrivacyRequestResumeFormat
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import (
    BulkPostPrivacyRequests,
    BulkReviewResponse,
    BulkSoftDeletePrivacyRequests,
    CheckpointActionRequired,
    DenyPrivacyRequests,
    ExecutionLogDetailResponse,
    ExecutionLogStatus,
    FilteredPrivacyRequestResults,
    LogEntry,
    ManualWebhookData,
    PrivacyRequestAccessResults,
    PrivacyRequestCreate,
    PrivacyRequestFilter,
    PrivacyRequestNotificationInfo,
    PrivacyRequestResponse,
    PrivacyRequestSource,
    PrivacyRequestStatus,
    PrivacyRequestTaskSchema,
    PrivacyRequestVerboseResponse,
    RequestTaskCallbackRequest,
    ReviewPrivacyRequestIds,
    VerificationCode,
)
from fides.api.service.deps import get_messaging_service, get_privacy_request_service
from fides.api.service.messaging.message_dispatch_service import EMAIL_JOIN_STRING
from fides.api.task.execute_request_tasks import log_task_queued, queue_request_task
from fides.api.task.filter_results import filter_data_categories
from fides.api.task.graph_task import EMPTY_REQUEST, EMPTY_REQUEST_TASK, collect_queries
from fides.api.task.task_resources import TaskResources
from fides.api.util.api_router import APIRouter
from fides.api.util.cache import FidesopsRedis, get_cache
from fides.api.util.collection_util import Row
from fides.api.util.endpoint_utils import validate_start_and_end_filters
from fides.api.util.enums import ColumnSort
from fides.api.util.fuzzy_search_utils import get_decrypted_identities_automaton
from fides.api.util.storage_util import storage_json_encoder
from fides.common.api.scope_registry import (
    PRIVACY_REQUEST_CALLBACK_RESUME,
    PRIVACY_REQUEST_CREATE,
    PRIVACY_REQUEST_DELETE,
    PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE,
    PRIVACY_REQUEST_NOTIFICATIONS_READ,
    PRIVACY_REQUEST_READ,
    PRIVACY_REQUEST_READ_ACCESS_RESULTS,
    PRIVACY_REQUEST_REVIEW,
    PRIVACY_REQUEST_TRANSFER,
    PRIVACY_REQUEST_UPLOAD_DATA,
    PRIVACY_REQUEST_VIEW_DATA,
)
from fides.common.api.v1.urn_registry import (
    PRIVACY_REQUEST_ACCESS_RESULTS,
    PRIVACY_REQUEST_APPROVE,
    PRIVACY_REQUEST_AUTHENTICATED,
    PRIVACY_REQUEST_BULK_RETRY,
    PRIVACY_REQUEST_BULK_SOFT_DELETE,
    PRIVACY_REQUEST_DENY,
    PRIVACY_REQUEST_FILTERED_RESULTS,
    PRIVACY_REQUEST_MANUAL_WEBHOOK_ACCESS_INPUT,
    PRIVACY_REQUEST_MANUAL_WEBHOOK_ERASURE_INPUT,
    PRIVACY_REQUEST_NOTIFICATIONS,
    PRIVACY_REQUEST_PRE_APPROVE_ELIGIBLE,
    PRIVACY_REQUEST_PRE_APPROVE_NOT_ELIGIBLE,
    PRIVACY_REQUEST_REQUEUE,
    PRIVACY_REQUEST_RESUBMIT,
    PRIVACY_REQUEST_RESUME,
    PRIVACY_REQUEST_RESUME_FROM_REQUIRES_INPUT,
    PRIVACY_REQUEST_RETRY,
    PRIVACY_REQUEST_SEARCH,
    PRIVACY_REQUEST_SOFT_DELETE,
    PRIVACY_REQUEST_TRANSFER_TO_PARENT,
    PRIVACY_REQUEST_VERIFY_IDENTITY,
    PRIVACY_REQUESTS,
    REQUEST_PREVIEW,
    REQUEST_STATUS_LOGS,
    REQUEST_TASK_CALLBACK,
    REQUEST_TASKS,
    V1_URL_PREFIX,
)
from fides.config import CONFIG
from fides.config.config_proxy import ConfigProxy
from fides.service.dataset.dataset_config_service import (
    replace_references_with_identities,
)
from fides.service.messaging.messaging_service import MessagingService
from fides.service.privacy_request.privacy_request_service import (
    PrivacyRequestError,
    PrivacyRequestService,
    _process_privacy_request_restart,
    _requeue_privacy_request,
    handle_approval,
    queue_privacy_request,
)

router = APIRouter(tags=["Privacy Requests"], prefix=V1_URL_PREFIX)

EMBEDDED_EXECUTION_LOG_LIMIT = 50


def get_privacy_request_or_error(
    db: Session, privacy_request_id: str, error_if_deleted: Optional[bool] = True
) -> PrivacyRequest:
    """Load the privacy request or throw a 404"""
    logger.debug("Finding privacy request with id '{}'", privacy_request_id)

    privacy_request = PrivacyRequest.get(db, object_id=privacy_request_id)

    if not privacy_request:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No privacy request found with id '{privacy_request_id}'.",
        )

    if error_if_deleted and privacy_request.deleted_at is not None:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Privacy request with id {privacy_request_id} has been deleted.",
        )

    return privacy_request


@router.post(
    PRIVACY_REQUESTS,
    status_code=HTTP_200_OK,
    response_model=BulkPostPrivacyRequests,
)
def create_privacy_request(
    *,
    privacy_request_service: PrivacyRequestService = Depends(
        get_privacy_request_service
    ),
    data: Annotated[List[PrivacyRequestCreate], Field(max_length=50)],  # type: ignore
) -> BulkPostPrivacyRequests:
    """
    Given a list of privacy request data elements, create corresponding PrivacyRequest objects
    or report failure and execute them within the Fidesops system.
    You cannot update privacy requests after they've been created.
    """
    return privacy_request_service.create_bulk_privacy_requests(
        data, authenticated=False
    )


@router.post(
    PRIVACY_REQUEST_AUTHENTICATED,
    status_code=HTTP_200_OK,
    response_model=BulkPostPrivacyRequests,
)
def create_privacy_request_authenticated(
    *,
    privacy_request_service: PrivacyRequestService = Depends(
        get_privacy_request_service
    ),
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[PRIVACY_REQUEST_CREATE],
    ),
    data: Annotated[List[PrivacyRequestCreate], Field(max_length=50)],  # type: ignore
) -> BulkPostPrivacyRequests:
    """
    Given a list of privacy request data elements, create corresponding PrivacyRequest objects
    or report failure and execute them within the Fidesops system.
    You cannot update privacy requests after they've been created.
    This route requires authentication instead of using verification codes.
    """

    return privacy_request_service.create_bulk_privacy_requests(
        data, authenticated=True, user_id=client.user_id
    )


def privacy_request_csv_download(
    db: Session, privacy_request_query: Query
) -> StreamingResponse:
    """Download privacy requests as CSV for Admin UI"""
    f = io.StringIO()
    csv_file = csv.writer(f)

    csv_file.writerow(
        [
            "Status",
            "Request Type",
            "Subject Identity",
            "Custom Privacy Request Fields",
            "Time Received",
            "Reviewed By",
            "Request ID",
            "Time Approved/Denied",
            "Denial Reason",
        ]
    )
    privacy_request_ids: List[str] = [r.id for r in privacy_request_query]
    denial_audit_log_query: Query = db.query(AuditLog).filter(
        AuditLog.action == AuditLogAction.denied,
        AuditLog.privacy_request_id.in_(privacy_request_ids),
    )
    denial_audit_logs: Dict[str, str] = {
        r.privacy_request_id: r.message for r in denial_audit_log_query
    }

    for pr in privacy_request_query:
        denial_reason = (
            denial_audit_logs[pr.id]
            if pr.status == PrivacyRequestStatus.denied and pr.id in denial_audit_logs
            else None
        )

        csv_file.writerow(
            [
                pr.status.value if pr.status else None,
                pr.policy.rules[0].action_type if len(pr.policy.rules) > 0 else None,
                pr.get_persisted_identity().model_dump(mode="json"),
                pr.get_persisted_custom_privacy_request_fields(),
                pr.created_at,
                pr.reviewed_by,
                pr.id,
                pr.reviewed_at,
                denial_reason,
            ]
        )

    f.seek(0)
    response = StreamingResponse(f, media_type="text/csv")
    response.headers["Content-Disposition"] = (
        f"attachment; filename=privacy_requests_download_{datetime.today().strftime('%Y-%m-%d')}.csv"
    )
    return response


def execution_and_audit_logs_by_dataset_name(
    self: PrivacyRequest,
) -> DefaultDict[str, List[Union["AuditLog", "ExecutionLog"]]]:
    """
    Returns a combined mapping of execution and audit logs for the given privacy request.

    Audit Logs are for the entire privacy request as a whole, while execution logs are created for specific collections.
    Logs here are grouped by dataset, but if it is an audit log, it is just given a fake dataset name, here "Request + status"
    ExecutionLogs for each dataset are truncated.

    Added as a conditional property to the PrivacyRequest class at runtime to
    show optionally embedded execution and audit logs.

    An example response might include your execution logs from your mongo db in one group, and execution logs from
    your postgres db in a different group, plus audit logs for when the request was approved and denied.
    """
    db: Session = Session.object_session(self)
    all_logs: DefaultDict[str, List[Union["AuditLog", "ExecutionLog"]]] = defaultdict(
        list
    )

    execution_log_query: Query = db.query(
        ExecutionLog.id,
        ExecutionLog.created_at,
        ExecutionLog.updated_at,
        ExecutionLog.message,
        cast(ExecutionLog.status, sqlalchemy.String).label(
            "status"
        ),  # Casting to string so we can perform a union of execution log and audit log statuses
        ExecutionLog.privacy_request_id,
        ExecutionLog.dataset_name,
        ExecutionLog.collection_name,
        ExecutionLog.connection_key,
        ExecutionLog.fields_affected,
        ExecutionLog.action_type,
        null().label("user_id"),
    ).filter(ExecutionLog.privacy_request_id == self.id)

    audit_log_query: Query = db.query(
        AuditLog.id,
        AuditLog.created_at,
        AuditLog.updated_at,
        AuditLog.message,
        cast(AuditLog.action.label("status"), sqlalchemy.String).label(
            "status"
        ),  # Casting to string so we can perform a union of execution log and audit log statuses
        AuditLog.privacy_request_id,
        null().label("dataset_name"),
        null().label("collection_name"),
        null().label("connection_key"),
        null().label("fields_affected"),
        null().label("action_type"),
        AuditLog.user_id,
    ).filter(AuditLog.privacy_request_id == self.id)

    combined: Query = execution_log_query.union_all(audit_log_query)

    for log in combined.order_by(ExecutionLog.updated_at.asc()):
        dataset_name: str = log.dataset_name or f"Request {log.status}"

        if len(all_logs[dataset_name]) > EMBEDDED_EXECUTION_LOG_LIMIT - 1:
            continue
        all_logs[dataset_name].append(log)
    return all_logs


def _filter_privacy_request_queryset(
    db: Session,
    query: Query,
    request_id: Optional[str] = None,
    identity: Optional[str] = None,
    fuzzy_search_str: Optional[str] = None,
    identities: Optional[Dict[str, Any]] = None,
    custom_privacy_request_fields: Optional[Dict[str, Any]] = None,
    status: Optional[List[PrivacyRequestStatus]] = None,
    created_lt: Optional[datetime] = None,
    created_gt: Optional[datetime] = None,
    started_lt: Optional[datetime] = None,
    started_gt: Optional[datetime] = None,
    completed_lt: Optional[datetime] = None,
    completed_gt: Optional[datetime] = None,
    errored_lt: Optional[datetime] = None,
    errored_gt: Optional[datetime] = None,
    external_id: Optional[str] = None,
    action_type: Optional[ActionType] = None,
    include_consent_webhook_requests: Optional[bool] = False,
    include_deleted_requests: Optional[bool] = False,
) -> Query:
    """
    Utility method to apply filters to our privacy request query.

    Status supports "or" filtering:
    `status=["approved","pending"]` will be translated into an "or" query.

    The `identities` and `custom_privacy_request_fields` parameters allow
    searching for privacy requests that match any of the provided identities
    or custom privacy request fields, respectively. The filtering is performed
    using an "or" condition, meaning that a privacy request will be included
    in the results if it matches at least one of the provided identities or
    custom privacy request fields.
    """

    if any([completed_lt, completed_gt]) and any([errored_lt, errored_gt]):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Cannot specify both succeeded and failed query params.",
        )

    validate_start_and_end_filters(
        [
            (created_lt, created_gt, "created"),
            (completed_lt, completed_gt, "completed"),
            (errored_lt, errored_gt, "errored"),
            (started_lt, started_gt, "started"),
        ]
    )

    # Handle fuzzy search string
    if fuzzy_search_str:
        if CONFIG.execution.fuzzy_search_enabled:
            decrypted_identities_automaton = get_decrypted_identities_automaton(db)

            # Set of associated privacy request ids
            fuzzy_search_identity_privacy_request_ids: Optional[Set[str]] = set(
                x
                for list in decrypted_identities_automaton.values(fuzzy_search_str)
                for x in list
            )

            if not fuzzy_search_identity_privacy_request_ids:
                query = query.filter(PrivacyRequest.id.ilike(f"{fuzzy_search_str}%"))
            else:
                query = query.filter(
                    or_(
                        PrivacyRequest.id.in_(
                            fuzzy_search_identity_privacy_request_ids
                        ),
                        PrivacyRequest.id.ilike(f"{fuzzy_search_str}%"),
                    )
                )
        else:
            # When fuzzy search is disabled, treat fuzzy_search_str as an
            # exact match on identity or partial match on privacy request ID
            identity_hashes = ProvidedIdentity.hash_value_for_search(fuzzy_search_str)
            identity_set: Set[str] = {
                identity[0]
                for identity in ProvidedIdentity.filter(
                    db=db,
                    conditions=(
                        (ProvidedIdentity.hashed_value.in_(identity_hashes))
                        & (ProvidedIdentity.privacy_request_id.isnot(None))
                    ),
                ).values(column("privacy_request_id"))
            }

            query = query.filter(
                or_(
                    PrivacyRequest.id.in_(identity_set),
                    PrivacyRequest.id.ilike(f"%{fuzzy_search_str}%"),
                )
            )

    if identity:
        identity_hashes = ProvidedIdentity.hash_value_for_search(identity)
        identity_set: Set[str] = {  # type: ignore[no-redef]
            identity[0]
            for identity in ProvidedIdentity.filter(
                db=db,
                conditions=(
                    (ProvidedIdentity.hashed_value.in_(identity_hashes))
                    & (ProvidedIdentity.privacy_request_id.isnot(None))
                ),
            ).values(column("privacy_request_id"))
        }
        query = query.filter(PrivacyRequest.id.in_(identity_set))

    if identities:
        identity_conditions = [
            (ProvidedIdentity.field_name == field_name)
            & (
                ProvidedIdentity.hashed_value.in_(
                    ProvidedIdentity.hash_value_for_search(value)
                )
            )
            for field_name, value in identities.items()
        ]

        identities_query = select([ProvidedIdentity.privacy_request_id]).where(
            or_(*identity_conditions)
            & (ProvidedIdentity.privacy_request_id.isnot(None))
        )
        query = query.filter(PrivacyRequest.id.in_(identities_query))

    if custom_privacy_request_fields:
        # Note that if Custom Privacy Request Field values were arrays, they are not
        # indexed for searching and not discoverable here
        custom_field_conditions = [
            (CustomPrivacyRequestField.field_name == field_name)
            & (
                CustomPrivacyRequestField.hashed_value.in_(
                    CustomPrivacyRequestField.hash_value_for_search(value)
                )
            )
            for field_name, value in custom_privacy_request_fields.items()
            if not isinstance(value, list)
        ]

        custom_fields_query = select(
            [CustomPrivacyRequestField.privacy_request_id]
        ).where(
            or_(*custom_field_conditions)
            & (CustomPrivacyRequestField.privacy_request_id.isnot(None))
        )
        query = query.filter(PrivacyRequest.id.in_(custom_fields_query))

    # Further restrict all PrivacyRequests by additional params
    if request_id:
        query = query.filter(PrivacyRequest.id.ilike(f"%{request_id}%"))
    if external_id:
        query = query.filter(PrivacyRequest.external_id.ilike(f"{external_id}%"))
    if status:
        query = query.filter(PrivacyRequest.status.in_(status))
    if created_lt:
        query = query.filter(PrivacyRequest.created_at < created_lt)
    if created_gt:
        query = query.filter(PrivacyRequest.created_at > created_gt)
    if started_lt:
        query = query.filter(PrivacyRequest.started_processing_at < started_lt)
    if started_gt:
        query = query.filter(PrivacyRequest.started_processing_at > started_gt)
    if completed_lt:
        query = query.filter(
            PrivacyRequest.status == PrivacyRequestStatus.complete,
            PrivacyRequest.finished_processing_at < completed_lt,
        )
    if completed_gt:
        query = query.filter(
            PrivacyRequest.status == PrivacyRequestStatus.complete,
            PrivacyRequest.finished_processing_at > completed_gt,
        )
    if errored_lt:
        query = query.filter(
            PrivacyRequest.status == PrivacyRequestStatus.error,
            PrivacyRequest.finished_processing_at < errored_lt,
        )
    if errored_gt:
        query = query.filter(
            PrivacyRequest.status == PrivacyRequestStatus.error,
            PrivacyRequest.finished_processing_at > errored_gt,
        )
    if action_type:
        policy_ids_for_action_type = (
            db.query(Rule)
            .filter(Rule.action_type == action_type)
            .with_entities(Rule.policy_id)
            .distinct()
        )
        query = query.filter(PrivacyRequest.policy_id.in_(policy_ids_for_action_type))

    if not include_consent_webhook_requests:
        query = query.filter(
            or_(
                PrivacyRequest.source != PrivacyRequestSource.consent_webhook,
                PrivacyRequest.source.is_(None),
            )
        )

    # Filter out test privacy requests
    query = query.filter(
        or_(
            PrivacyRequest.source != PrivacyRequestSource.dataset_test,
            PrivacyRequest.source.is_(None),
        )
    )

    # Filter out deleted requests
    if not include_deleted_requests:
        query = query.filter(PrivacyRequest.deleted_at.is_(None))

    return query


def _sort_privacy_request_queryset(
    query: Query, sort_field: str, sort_direction: ColumnSort
) -> Query:
    if hasattr(PrivacyRequest, sort_field) is False:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"{sort_field} is not on PrivacyRequest",
        )

    sort_object_attribute = getattr(PrivacyRequest, sort_field)
    sort_func = getattr(sort_object_attribute, sort_direction)
    return query.order_by(nullslast(sort_func()))


def attach_resume_instructions(privacy_request: PrivacyRequest) -> None:
    """
    Temporarily update a paused/errored/requires_input privacy request object with instructions from the Redis cache
    about how to resume manually if applicable.
    """
    resume_endpoint: Optional[str] = None
    action_required_details: Optional[CheckpointActionRequired] = None

    if privacy_request.status == PrivacyRequestStatus.paused:
        # Graph is paused on a pre-processing webhook
        resume_endpoint = PRIVACY_REQUEST_RESUME

    elif privacy_request.status == PrivacyRequestStatus.error:
        action_required_details = privacy_request.get_failed_checkpoint_details()
        resume_endpoint = PRIVACY_REQUEST_RETRY

    elif privacy_request.status == PrivacyRequestStatus.requires_input:
        # No action required details because this doesn't need to resume from a
        # specific step or collection
        resume_endpoint = PRIVACY_REQUEST_RESUME_FROM_REQUIRES_INPUT

    if action_required_details:
        action_required_details.step = action_required_details.step.value  # type: ignore
        action_required_details.collection = (
            action_required_details.collection.value if action_required_details.collection else None  # type: ignore
        )

    privacy_request.action_required_details = action_required_details
    # replaces the placeholder in the url with the privacy request id
    privacy_request.resume_endpoint = (
        resume_endpoint.format(privacy_request_id=privacy_request.id)
        if resume_endpoint
        else None
    )


def _validate_result_size(query: Query) -> None:
    """
    Validates the result size is less than maximum allowed by settings.
    Raises an HTTPException if result size is greater than maximum.
    Result size is determined by running an up-front "count" query.
    """
    row_count = query.count()
    max_rows = CONFIG.admin_ui.max_privacy_request_download_rows
    if row_count > max_rows:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Requested privacy request report would contain {row_count} rows. A maximum of {max_rows} rows is permitted. Please narrow your date range and try again.",
        )


def _shared_privacy_request_search(
    *,
    db: Session,
    params: Params,
    request_id: Optional[str] = None,
    identity: Optional[str] = None,
    fuzzy_search_str: Optional[str] = None,
    identities: Optional[Dict[str, str]] = None,
    custom_privacy_request_fields: Optional[Dict[str, Any]] = None,
    status: Optional[List[PrivacyRequestStatus]] = None,
    created_lt: Optional[datetime] = None,
    created_gt: Optional[datetime] = None,
    started_lt: Optional[datetime] = None,
    started_gt: Optional[datetime] = None,
    completed_lt: Optional[datetime] = None,
    completed_gt: Optional[datetime] = None,
    errored_lt: Optional[datetime] = None,
    errored_gt: Optional[datetime] = None,
    external_id: Optional[str] = None,
    action_type: Optional[ActionType] = None,
    verbose: Optional[bool] = False,
    include_identities: Optional[bool] = False,
    include_custom_privacy_request_fields: Optional[bool] = False,
    include_deleted_requests: Optional[bool] = False,
    download_csv: Optional[bool] = False,
    sort_field: str = "created_at",
    sort_direction: ColumnSort = ColumnSort.DESC,
) -> Union[StreamingResponse, AbstractPage[PrivacyRequest]]:
    """
    Internal function to handle the logic for retrieving privacy requests.

    This function is used by both the GET and POST versions of the privacy request endpoints
    to avoid duplicating the logic while transitioning from the GET version to the
    POST version of the endpoint.
    """

    logger.debug("Finding all request statuses with pagination params {}", params)

    query = db.query(PrivacyRequest)
    query = _filter_privacy_request_queryset(
        db,
        query,
        request_id,
        identity,
        fuzzy_search_str,
        identities,
        custom_privacy_request_fields,
        status,
        created_lt,
        created_gt,
        started_lt,
        started_gt,
        completed_lt,
        completed_gt,
        errored_lt,
        errored_gt,
        external_id,
        action_type,
        None,
        include_deleted_requests,
    )

    logger.debug(
        "Sorting requests by field: {} and direction: {}", sort_field, sort_direction
    )
    query = _sort_privacy_request_queryset(query, sort_field, sort_direction)

    if download_csv:
        _validate_result_size(query)
        # Returning here if download_csv param was specified
        logger.info("Downloading privacy requests as csv")
        return privacy_request_csv_download(db, query)

    # Conditionally embed execution log details in the response.
    if verbose:
        logger.info("Finding execution and audit log details")
        PrivacyRequest.execution_and_audit_logs_by_dataset = property(
            execution_and_audit_logs_by_dataset_name
        )
    else:
        PrivacyRequest.execution_and_audit_logs_by_dataset = property(lambda self: None)

    paginated = paginate(query, params)

    for item in paginated.items:  # type: ignore
        if include_identities:
            item.identity = item.get_persisted_identity().labeled_dict(
                include_default_labels=True
            )

        if include_custom_privacy_request_fields:
            item.custom_privacy_request_fields = (
                item.get_persisted_custom_privacy_request_fields()
            )

        attach_resume_instructions(item)

    return paginated


@router.get(
    PRIVACY_REQUESTS,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ])],
    response_model=Page[
        Union[
            PrivacyRequestVerboseResponse,
            PrivacyRequestResponse,
        ]
    ],
    deprecated=True,
)
def get_request_status(
    *,
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
    request_id: Optional[str] = None,
    identity: Optional[str] = None,
    status: Optional[List[PrivacyRequestStatus]] = FastAPIQuery(
        default=None
    ),  # type:ignore
    fuzzy_search_str: Optional[str] = None,
    created_lt: Optional[datetime] = None,
    created_gt: Optional[datetime] = None,
    started_lt: Optional[datetime] = None,
    started_gt: Optional[datetime] = None,
    completed_lt: Optional[datetime] = None,
    completed_gt: Optional[datetime] = None,
    errored_lt: Optional[datetime] = None,
    errored_gt: Optional[datetime] = None,
    external_id: Optional[str] = None,
    action_type: Optional[ActionType] = None,
    verbose: Optional[bool] = False,
    include_identities: Optional[bool] = False,
    include_custom_privacy_request_fields: Optional[bool] = False,
    download_csv: Optional[bool] = False,
    include_deleted_requests: Optional[bool] = False,
    sort_field: str = "created_at",
    sort_direction: ColumnSort = ColumnSort.DESC,
) -> Union[StreamingResponse, AbstractPage[PrivacyRequest]]:
    """
    **This endpoint is deprecated. Please use `POST /privacy-request/search`,
    which uses body parameters instead of query parameters for filtering.**

    Returns PrivacyRequest information. Supports a variety of optional query params.

    To fetch a single privacy request, use the request_id query param `?request_id=`.
    To see individual execution logs, use the verbose query param `?verbose=True`.
    """

    # Both the old and new versions of the privacy request search endpoints use this shared function.
    # The `identities` and `custom_privacy_request_fields` parameters are only supported in the new version
    # so they are both set to None here.
    return _shared_privacy_request_search(
        db=db,
        params=params,
        request_id=request_id,
        identity=identity,
        fuzzy_search_str=fuzzy_search_str,
        identities=None,
        custom_privacy_request_fields=None,
        status=status,
        created_lt=created_lt,
        created_gt=created_gt,
        started_lt=started_lt,
        started_gt=started_gt,
        completed_lt=completed_lt,
        completed_gt=completed_gt,
        errored_lt=errored_lt,
        errored_gt=errored_gt,
        external_id=external_id,
        action_type=action_type,
        verbose=verbose,
        include_identities=include_identities,
        include_custom_privacy_request_fields=include_custom_privacy_request_fields,
        include_deleted_requests=include_deleted_requests,
        download_csv=download_csv,
        sort_field=sort_field,
        sort_direction=sort_direction,
    )


@router.post(
    PRIVACY_REQUEST_SEARCH,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ])],
    response_model=Page[
        Union[
            PrivacyRequestVerboseResponse,
            PrivacyRequestResponse,
        ]
    ],
)
def privacy_request_search(
    *,
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
    privacy_request_filter: Optional[PrivacyRequestFilter] = Body(None),
) -> Union[StreamingResponse, AbstractPage[PrivacyRequest]]:
    """
    Returns PrivacyRequest information. Supports a variety of optional filter parameters.

    To see individual execution logs, set `"verbose": true`.
    """

    # default filter if the payload is empty
    if privacy_request_filter is None:
        privacy_request_filter = PrivacyRequestFilter()

    return _shared_privacy_request_search(
        db=db,
        params=params,
        request_id=privacy_request_filter.request_id,
        identities=privacy_request_filter.identities,
        fuzzy_search_str=privacy_request_filter.fuzzy_search_str,
        custom_privacy_request_fields=privacy_request_filter.custom_privacy_request_fields,
        status=privacy_request_filter.status,  # type: ignore
        created_lt=privacy_request_filter.created_lt,
        created_gt=privacy_request_filter.created_gt,
        started_lt=privacy_request_filter.started_lt,
        started_gt=privacy_request_filter.started_gt,
        completed_lt=privacy_request_filter.completed_lt,
        completed_gt=privacy_request_filter.completed_gt,
        errored_lt=privacy_request_filter.errored_lt,
        errored_gt=privacy_request_filter.errored_gt,
        external_id=privacy_request_filter.external_id,
        action_type=privacy_request_filter.action_type,
        verbose=privacy_request_filter.verbose,
        include_identities=privacy_request_filter.include_identities,
        include_custom_privacy_request_fields=privacy_request_filter.include_custom_privacy_request_fields,
        include_deleted_requests=privacy_request_filter.include_deleted_requests,
        download_csv=privacy_request_filter.download_csv,
        sort_field=privacy_request_filter.sort_field,
        sort_direction=privacy_request_filter.sort_direction,
    )


@router.get(
    REQUEST_STATUS_LOGS,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ])],
    response_model=Page[ExecutionLogDetailResponse],
)
def get_request_status_logs(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
) -> AbstractPage[ExecutionLog]:
    """Returns all the execution logs associated with a given privacy request ordered by updated asc."""

    get_privacy_request_or_error(db, privacy_request_id, error_if_deleted=False)

    logger.debug(
        "Finding all execution logs for privacy request {} with params '{}'",
        privacy_request_id,
        params,
    )

    return paginate(
        ExecutionLog.query(db=db)
        .filter(ExecutionLog.privacy_request_id == privacy_request_id)
        .order_by(ExecutionLog.updated_at.asc()),
        params,
    )


@router.get(
    PRIVACY_REQUEST_NOTIFICATIONS,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestNotificationInfo,
    dependencies=[
        Security(
            verify_oauth_client,
            scopes=[PRIVACY_REQUEST_NOTIFICATIONS_READ],
        )
    ],
)
def get_privacy_request_notification_info(
    *, db: Session = Depends(deps.get_db)
) -> PrivacyRequestNotificationInfo:
    """Retrieve privacy request notification email addresses and number of failures to trigger notifications."""
    info = PrivacyRequestNotifications.all(db)

    if not info:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No privacy request notification info found",
        )

    return PrivacyRequestNotificationInfo(
        email_addresses=[x for x in info[0].email.split(EMAIL_JOIN_STRING)],
        notify_after_failures=info[0].notify_after_failures,
    )


@router.put(
    PRIVACY_REQUEST_NOTIFICATIONS,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestNotificationInfo,
    dependencies=[
        Security(
            verify_oauth_client,
            scopes=[PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE],
        )
    ],
)
def create_or_update_privacy_request_notifications(
    *, db: Session = Depends(deps.get_db), request_body: PrivacyRequestNotificationInfo
) -> PrivacyRequestNotificationInfo:
    """Create or update list of email addresses and number of failures for privacy request notifications."""
    # If email_addresses is empty it means notifications were turned off and the email
    # information should be deleted from the database. In this situation an empty list
    # of email address is returned along with the notify_after_failures sent from the
    # front end. This allows the first end to control the default notifiy_after_failures
    # number.
    if not request_body.email_addresses:
        all_notes = PrivacyRequestNotifications.all(db)
        try:
            note: PrivacyRequestNotifications = all_notes[0]
        except IndexError:
            pass
        else:
            note.delete(db=db)

        return PrivacyRequestNotificationInfo(
            email_addresses=[],
            notify_after_failures=request_body.notify_after_failures,
        )

    notification_info = {
        "email": EMAIL_JOIN_STRING.join(request_body.email_addresses),
        "notify_after_failures": request_body.notify_after_failures,
    }
    info_check: List[PrivacyRequestNotifications] = PrivacyRequestNotifications.all(
        db=db
    )
    info: PrivacyRequestNotifications
    try:
        info = info_check[0]
    except IndexError:
        info = PrivacyRequestNotifications.create(
            db=db,
            data=notification_info,
        )
    else:
        info.update(
            db=db,
            data=notification_info,
        )

    return PrivacyRequestNotificationInfo(
        email_addresses=info.email.split(", "),
        notify_after_failures=info.notify_after_failures,
    )


@router.put(
    REQUEST_PREVIEW,
    status_code=HTTP_200_OK,
    response_model=List[DryRunDatasetResponse],
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ])],
)
def get_request_preview_queries(
    *,
    db: Session = Depends(deps.get_db),
    dataset_keys: Optional[List[str]] = Body(None),
) -> List[DryRunDatasetResponse]:
    """Returns dry run queries given a list of dataset ids.  If a dataset references another dataset, both dataset
    keys must be in the request body."""
    dataset_configs: List[DatasetConfig] = []
    if not dataset_keys:
        dataset_configs = DatasetConfig.all(db=db)
        if not dataset_configs:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND,
                detail="No datasets could be found",
            )
    else:
        for dataset_key in dataset_keys:
            dataset_config = DatasetConfig.get_by(
                db=db, field="fides_key", value=dataset_key
            )
            if not dataset_config:
                raise HTTPException(
                    status_code=HTTP_404_NOT_FOUND,
                    detail=f"No dataset with id '{dataset_key}'",
                )
            dataset_configs.append(dataset_config)
    try:
        connection_configs: List[ConnectionConfig] = []
        for dataset in dataset_configs:
            connection_config: Optional[ConnectionConfig] = ConnectionConfig.get(
                db=db, object_id=dataset.connection_config_id
            )
            if connection_config:
                connection_configs.append(connection_config)

        try:
            dataset_graph: DatasetGraph = DatasetGraph(
                *[dataset.get_graph() for dataset in dataset_configs]
            )
        except ValidationError as exc:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=f"{exc}. Make sure all referenced datasets are included in the request body.",
            )

        identity_seed: Dict[str, str] = {
            k: "something" for k in dataset_graph.identity_keys.values()
        }
        traversal: Traversal = Traversal(dataset_graph, identity_seed)

        queries: Dict[CollectionAddress, str] = collect_queries(
            traversal,
            TaskResources(
                EMPTY_REQUEST, Policy(), connection_configs, EMPTY_REQUEST_TASK, db
            ),
        )
        return [
            DryRunDatasetResponse(
                collectionAddress=CollectionAddressResponse(
                    dataset=key.dataset, collection=key.collection
                ),
                query=value,
            )
            for key, value in queries.items()
        ]
    except TraversalError as err:
        logger.info("Dry run failed: {}", err)
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Dry run failed",
        )


@router.post(
    PRIVACY_REQUEST_RESUME,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestResponse,
)
def resume_privacy_request(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    webhook: PolicyPreWebhook = Security(
        verify_callback_oauth_policy_pre_webhook,
        scopes=[PRIVACY_REQUEST_CALLBACK_RESUME],
    ),
    webhook_callback: PrivacyRequestResumeFormat,
) -> PrivacyRequestResponse:
    """Resume running a privacy request after it was paused by a Pre-Execution webhook"""
    privacy_request = get_privacy_request_or_error(db, privacy_request_id)
    # We don't want to persist derived identities because they have not been provided
    # by the end user
    privacy_request.cache_identity(webhook_callback.derived_identity)  # type: ignore

    if privacy_request.status != PrivacyRequestStatus.paused:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Invalid resume request: privacy request '{privacy_request.id}' status = {privacy_request.status.value}.",
            # type: ignore
        )

    logger.info(
        "Resuming privacy request '{}' from webhook '{}'",
        privacy_request_id,
        webhook.key,
    )

    privacy_request.status = PrivacyRequestStatus.in_processing
    privacy_request.save(db=db)

    queue_privacy_request(
        privacy_request_id=privacy_request.id,
        from_webhook_id=webhook.id,
    )
    return privacy_request  # type: ignore[return-value]


def validate_manual_input(
    manual_rows: List[Row],
    collection: CollectionAddress,
    dataset_graph: DatasetGraph,
) -> None:
    """Validate manually-added data for a collection.

    The specified collection must exist and all fields must be previously defined.
    """
    for row in manual_rows:
        for field_name in row:
            if not dataset_graph.nodes[collection].collection.contains_field(
                lambda f: f.name == field_name  # pylint: disable=W0640
            ):
                raise HTTPException(
                    status_code=HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"Cannot save manual rows. No '{field_name}' field defined on the '{collection.value}' collection.",
                )


@router.post(
    PRIVACY_REQUEST_BULK_RETRY,
    status_code=HTTP_200_OK,
    response_model=BulkPostPrivacyRequests,
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_CALLBACK_RESUME])
    ],
)
def bulk_restart_privacy_request_from_failure(
    privacy_request_ids: List[str],
    *,
    db: Session = Depends(deps.get_db),
) -> BulkPostPrivacyRequests:
    """Bulk restart a of privacy request from failure."""
    succeeded: List[PrivacyRequestResponse] = []
    failed: List[Dict[str, Any]] = []
    for privacy_request_id in privacy_request_ids:
        privacy_request = PrivacyRequest.get(db, object_id=privacy_request_id)

        if not privacy_request:
            failed.append(
                {
                    "message": f"No privacy request found with id '{privacy_request_id}'",
                    "data": {"privacy_request_id": privacy_request_id},
                }
            )
            continue

        if privacy_request.deleted_at is not None:
            failed.append(
                {
                    "message": "Cannot restart a deleted privacy request",
                    "data": {"privacy_request_id": privacy_request_id},
                }
            )
            continue

        if privacy_request.status != PrivacyRequestStatus.error:
            failed.append(
                {
                    "message": f"Cannot restart privacy request from failure: privacy request '{privacy_request.id}' status = {privacy_request.status.value}.",
                    "data": {"privacy_request_id": privacy_request_id},
                }
            )
            continue

        failed_details: Optional[CheckpointActionRequired] = (
            privacy_request.get_failed_checkpoint_details()
        )

        succeeded.append(
            _process_privacy_request_restart(
                privacy_request,
                failed_details.step if failed_details else None,
                db,
            )
        )

    return BulkPostPrivacyRequests(succeeded=succeeded, failed=failed)


@router.post(
    PRIVACY_REQUEST_RETRY,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestResponse,
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_CALLBACK_RESUME])
    ],
)
def restart_privacy_request_from_failure(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    privacy_request_service: PrivacyRequestService = Depends(
        get_privacy_request_service
    ),
) -> PrivacyRequestResponse:
    """Restart a privacy request from failure"""
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id
    )

    # Automatically resubmit the request if the cache has expired
    if (
        not privacy_request.get_cached_identity_data()
        and privacy_request.status
        not in [PrivacyRequestStatus.complete, PrivacyRequestStatus.pending]
    ):
        logger.info(
            f"Cached data for privacy request {privacy_request.id} has expired, automatically resubmitting request"
        )
        return privacy_request_service.resubmit_privacy_request(privacy_request_id)  # type: ignore[return-value]

    if privacy_request.status != PrivacyRequestStatus.error:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Cannot restart privacy request from failure: privacy request '{privacy_request.id}' status = {privacy_request.status.value}.",
            # type: ignore
        )

    failed_details: Optional[CheckpointActionRequired] = (
        privacy_request.get_failed_checkpoint_details()
    )

    return _process_privacy_request_restart(
        privacy_request,
        failed_details.step if failed_details else None,
        db,
    )


@router.post(
    PRIVACY_REQUEST_VERIFY_IDENTITY,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestResponse,
)
def verify_identification_code(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    config_proxy: ConfigProxy = Depends(deps.get_config_proxy),
    messaging_service: MessagingService = Depends(get_messaging_service),
    provided_code: VerificationCode,
) -> PrivacyRequestResponse:
    """Verify the supplied identity verification code.

    If successful, and we don't need a separate manual request approval, queue the privacy request
    for execution.

    Fires pre-approval webhooks if configured.
    """

    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id
    )
    try:
        privacy_request.verify_identity(db, provided_code.code)
        policy: Optional[Policy] = Policy.get(
            db=db, object_id=privacy_request.policy_id
        )
        messaging_service.send_privacy_request_receipt(
            policy, privacy_request.get_persisted_identity(), privacy_request
        )

    except IdentityVerificationException as exc:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=exc.message)
    except PermissionError as exc:
        logger.info("Invalid verification code provided for {}.", privacy_request.id)
        raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail=exc.args[0])

    logger.info("Identity verified for {}.", privacy_request.id)

    handle_approval(db, config_proxy, privacy_request)

    return privacy_request  # type: ignore[return-value]


@router.patch(
    PRIVACY_REQUEST_APPROVE,
    status_code=HTTP_200_OK,
    response_model=BulkReviewResponse,
)
def approve_privacy_request(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[PRIVACY_REQUEST_REVIEW],
    ),
    privacy_request_service: PrivacyRequestService = Depends(
        get_privacy_request_service
    ),
    privacy_requests: ReviewPrivacyRequestIds,
) -> BulkReviewResponse:
    """Approve and dispatch a list of privacy requests and/or report failure"""

    return privacy_request_service.approve_privacy_requests(
        privacy_requests.request_ids, reviewed_by=client.user_id
    )


@router.patch(
    PRIVACY_REQUEST_DENY,
    status_code=HTTP_200_OK,
    response_model=BulkReviewResponse,
)
def deny_privacy_request(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[PRIVACY_REQUEST_REVIEW],
    ),
    privacy_request_service: PrivacyRequestService = Depends(
        get_privacy_request_service
    ),
    privacy_requests: DenyPrivacyRequests,
) -> BulkReviewResponse:
    """Deny a list of privacy requests and/or report failure"""

    return privacy_request_service.deny_privacy_requests(
        privacy_requests.request_ids, privacy_requests.reason, user_id=client.user_id
    )


def _validate_privacy_request_pending_or_error(
    db: Session, privacy_request_id: str
) -> None:
    privacy_request = PrivacyRequest.get(db, object_id=privacy_request_id)
    if not privacy_request:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No privacy request found with id: '{privacy_request_id}'.",
        )

    if privacy_request.status != PrivacyRequestStatus.pending:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Privacy request with id '{privacy_request_id}' is not in pending status.",
        )


@router.post(
    PRIVACY_REQUEST_PRE_APPROVE_ELIGIBLE,
    status_code=HTTP_200_OK,
)
def mark_privacy_request_pre_approve_eligible(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    privacy_request_service: PrivacyRequestService = Depends(
        get_privacy_request_service
    ),
    webhook: PreApprovalWebhook = Security(
        verify_callback_oauth_pre_approval_webhook, scopes=[PRIVACY_REQUEST_REVIEW]
    ),
) -> None:
    """
    Marks privacy request as eligible for automatic approval.
    If all webhook responses have been received and all are affirmative, proceed to queue privacy request
    """
    _validate_privacy_request_pending_or_error(db, privacy_request_id)
    logger.info(
        "Marking privacy request '{}' as eligible for automatic approval via webhook '{}' for connection config '{}'.",
        privacy_request_id,
        webhook.key,
        webhook.connection_config.key,
    )

    try:
        PreApprovalWebhookReply.create(
            db=db,
            data={
                "webhook_id": webhook.id,
                "privacy_request_id": privacy_request_id,
                "is_eligible": True,
            },
        )
    except Exception as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Failed to mark privacy request {privacy_request_id} as eligible due to: {str(exc)}",
        )

    all_webhooks = PreApprovalWebhook.all(db)

    def _reply_exists_for_webhook(webhook_id: str) -> bool:
        return bool(
            PreApprovalWebhookReply.filter(
                db=db,
                conditions=(
                    (PreApprovalWebhookReply.webhook_id == webhook_id)
                    & (PreApprovalWebhookReply.privacy_request_id == privacy_request_id)
                ),
            ).first()
        )

    # Check if not all replies have been received
    if not all(_reply_exists_for_webhook(webhook.id) for webhook in all_webhooks):
        # Subtlety here is that if a new webhook is configured between when the privacy request was created and when
        # the last webhook replied, then the privacy request can never be automatically approved. It must instead be
        # manually approved in the Admin UI.
        logger.info(
            "Not all pre-approval webhooks have responded for privacy request '{}'. Cannot automatically approve request.",
            privacy_request_id,
        )
        return

    # Check if all replies are true.  Reply ignored if its webhook has since been deleted.
    replies_for_privacy_request = (
        db.query(PreApprovalWebhookReply)
        .filter(
            PreApprovalWebhookReply.privacy_request_id == privacy_request_id,
            PreApprovalWebhookReply.webhook_id.isnot(None),
        )
        .all()
    )

    if not all(reply.is_eligible for reply in replies_for_privacy_request):
        logger.info(
            "Not all pre-approval webhooks have responded with eligible for privacy request '{}'. Cannot automatically approve request.",
            privacy_request_id,
        )
        return

    logger.info(
        "All pre-approval webhooks have responded with eligible for privacy request '{}'. Proceeding to automatically approve request.",
        privacy_request_id,
    )
    privacy_request_service.approve_privacy_requests(
        [privacy_request_id], webhook_id=webhook.id
    )


@router.post(
    PRIVACY_REQUEST_PRE_APPROVE_NOT_ELIGIBLE,
    status_code=HTTP_200_OK,
)
def mark_privacy_request_pre_approve_not_eligible(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    webhook: PreApprovalWebhook = Security(
        verify_callback_oauth_pre_approval_webhook, scopes=[PRIVACY_REQUEST_REVIEW]
    ),
) -> None:
    """Marks privacy request as not eligible for automatic approval, regardless of what other webhook responses we receive"""
    _validate_privacy_request_pending_or_error(db, privacy_request_id)
    logger.info(
        "Marking privacy request '{}' as not eligible for automatic approval via webhook '{}' for connection config '{}'.",
        privacy_request_id,
        webhook.key,
        webhook.connection_config.key,
    )
    try:
        PreApprovalWebhookReply.create(
            db=db,
            data={
                "webhook_id": webhook.id,
                "privacy_request_id": privacy_request_id,
                "is_eligible": False,
            },
        )
    except Exception as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Failed to mark privacy request {privacy_request_id} as not eligible due to: {str(exc)}",
        )


def _handle_manual_webhook_input(
    action: Literal["access", "erasure"],
    connection_config: ConnectionConfig,
    privacy_request_id: str,
    db: Session,
    input_data: Dict[str, Any],
) -> None:
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id
    )
    access_manual_webhook: AccessManualWebhook = get_access_manual_webhook_or_404(
        connection_config
    )

    if not privacy_request.status == PrivacyRequestStatus.requires_input:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Invalid manual webhook {action} upload request: privacy request '{privacy_request.id}' status = {privacy_request.status.value}.",
            # type: ignore
        )

    try:
        getattr(privacy_request, f"cache_manual_webhook_{action}_input")(
            access_manual_webhook, input_data
        )
    except PydanticValidationError as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(exc.errors(include_url=False, include_input=False)),
        )

    logger.info(
        "{} input saved for manual webhook '{}' for privacy_request '{}'.",
        action.capitalize(),
        access_manual_webhook,
        privacy_request,
    )


@router.patch(
    PRIVACY_REQUEST_MANUAL_WEBHOOK_ACCESS_INPUT,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_UPLOAD_DATA])],
    response_model=None,
)
def upload_manual_webhook_access_data(
    *,
    connection_config: ConnectionConfig = Depends(_get_connection_config),
    privacy_request_id: str,
    db: Session = Depends(deps.get_db),
    input_data: Dict[str, Any],
) -> None:
    """Upload manual access input for the privacy request for the fields defined on the access manual webhook.
    The data collected here is not included in the graph but uploaded directly to the user at the end
    of privacy request execution.

    Because a 'manual_webhook' ConnectionConfig has one AccessManualWebhook associated with it,
    we are using the ConnectionConfig key as the AccessManualWebhook identifier here.
    """
    _handle_manual_webhook_input(
        action="access",
        connection_config=connection_config,
        privacy_request_id=privacy_request_id,
        db=db,
        input_data=input_data,
    )


@router.patch(
    PRIVACY_REQUEST_MANUAL_WEBHOOK_ERASURE_INPUT,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_UPLOAD_DATA])],
    response_model=None,
)
def upload_manual_webhook_erasure_data(
    *,
    connection_config: ConnectionConfig = Depends(_get_connection_config),
    privacy_request_id: str,
    db: Session = Depends(deps.get_db),
    input_data: Dict[str, Any],
) -> None:
    """Upload manual erasure input for the privacy request for the fields defined on the access manual webhook.

    Because a 'manual_webhook' ConnectionConfig has one AccessManualWebhook associated with it,
    we are using the ConnectionConfig key as the AccessManualWebhook identifier here.
    """
    _handle_manual_webhook_input(
        action="erasure",
        connection_config=connection_config,
        privacy_request_id=privacy_request_id,
        db=db,
        input_data=input_data,
    )


@router.get(
    PRIVACY_REQUEST_TRANSFER_TO_PARENT,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_TRANSFER])],
    response_model=Dict[str, Optional[List[Row]]],
)
def privacy_request_data_transfer(
    *,
    privacy_request_id: str,
    rule_key: str,
    db: Session = Depends(deps.get_db),
    cache: FidesopsRedis = Depends(deps.get_cache),
) -> Dict[str, Optional[List[Row]]]:
    """Transfer access request information to the parent server."""
    privacy_request = PrivacyRequest.get(db=db, object_id=privacy_request_id)

    if not privacy_request:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No privacy request with id {privacy_request_id} found",
        )

    rule = Rule.filter(
        db=db, conditions=(Rule.key == rule_key)
    ).first()  # pylint: disable=superfluous-parens
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"Rule key {rule_key} not found",
        )

    value_dict: Dict[str, Optional[List[Row]]] = cache.get_encoded_objects_by_prefix(
        f"{privacy_request_id}__access_request"
    )

    if not value_dict:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No access request information found for privacy request id {privacy_request_id}",
        )

    access_result = {k.split("__")[-1]: v for k, v in value_dict.items()}
    datasets = DatasetConfig.all(db=db)
    if not datasets:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No datasets found for privacy request {privacy_request_id}",
        )

    dataset_graphs = [dataset_config.get_graph() for dataset_config in datasets]
    if not dataset_graphs:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset graphs found for privacy request {privacy_request_id}",
        )
    dataset_graph = DatasetGraph(*dataset_graphs)
    target_categories = {target.data_category for target in rule.targets}
    filtered_results: Optional[Dict[str, Optional[List[Row]]]] = filter_data_categories(
        access_result,  # type: ignore
        target_categories,
        dataset_graph,
    )

    if filtered_results is None:
        raise HTTPException(
            status_code=404,
            detail=f"No results found for privacy request {privacy_request_id}",
        )

    return filtered_results


@router.get(
    PRIVACY_REQUEST_MANUAL_WEBHOOK_ACCESS_INPUT,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_VIEW_DATA])],
    response_model=Optional[ManualWebhookData],
)
def view_uploaded_manual_webhook_data(
    *,
    connection_config: ConnectionConfig = Depends(_get_connection_config),
    privacy_request_id: str,
    db: Session = Depends(deps.get_db),
) -> Optional[ManualWebhookData]:
    """
    View uploaded data for this privacy request for the given access manual webhook

    If no data exists for this webhook, we just return all fields as None.
    If we have missing or extra fields saved, we'll just return the overlap between what is saved and what is defined on the webhook.

    If checked=False, data must be reviewed before submission. The privacy request should not be submitted as-is.
    """
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db,
        privacy_request_id,
        error_if_deleted=False,
    )
    access_manual_webhook: AccessManualWebhook = get_access_manual_webhook_or_404(
        connection_config
    )

    if not privacy_request.status == PrivacyRequestStatus.requires_input:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Invalid manual webhook access upload request: privacy request "
            f"'{privacy_request.id}' status = {privacy_request.status.value}.",  # type: ignore
        )

    try:
        logger.info(
            "Retrieving input data for access manual webhook '{}' for privacy request '{}'.",
            connection_config.key,
            privacy_request.id,
        )
        data: Dict[str, Any] = privacy_request.get_manual_webhook_access_input_strict(
            access_manual_webhook
        )
        checked = True
    except (
        PydanticValidationError,
        ManualWebhookFieldsUnset,
        NoCachedManualWebhookEntry,
    ) as exc:
        logger.info(exc)
        data = privacy_request.get_manual_webhook_access_input_non_strict(
            manual_webhook=access_manual_webhook
        )
        checked = False

    return ManualWebhookData(checked=checked, fields=data)


@router.get(
    PRIVACY_REQUEST_MANUAL_WEBHOOK_ERASURE_INPUT,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_VIEW_DATA])],
    response_model=Optional[ManualWebhookData],
)
def view_uploaded_erasure_manual_webhook_data(
    *,
    connection_config: ConnectionConfig = Depends(_get_connection_config),
    privacy_request_id: str,
    db: Session = Depends(deps.get_db),
) -> Optional[ManualWebhookData]:
    """
    View uploaded erasure data for this privacy request for the given manual webhook

    If no data exists for this webhook, we just return all fields as None.
    If we have missing or extra fields saved, we'll just return the overlap between what is saved and what is defined on the webhook.

    If checked=False, data must be reviewed before submission. The privacy request should not be submitted as-is.
    """
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db,
        privacy_request_id,
        error_if_deleted=False,
    )
    manual_webhook: AccessManualWebhook = get_access_manual_webhook_or_404(
        connection_config
    )

    if not privacy_request.status == PrivacyRequestStatus.requires_input:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Invalid manual webhook erasure upload request: privacy request "
            f"'{privacy_request.id}' status = {privacy_request.status.value}.",  # type: ignore
        )

    try:
        logger.info(
            "Retrieving erasure input data for manual webhook '{}' for privacy request '{}'.",
            connection_config.key,
            privacy_request.id,
        )
        data: Dict[str, Any] = privacy_request.get_manual_webhook_erasure_input_strict(
            manual_webhook
        )
        checked = True
    except (
        PydanticValidationError,
        ManualWebhookFieldsUnset,
        NoCachedManualWebhookEntry,
    ) as exc:
        logger.info(exc)
        data = privacy_request.get_manual_webhook_erasure_input_non_strict(
            manual_webhook=manual_webhook
        )
        checked = False

    return ManualWebhookData(checked=checked, fields=data)


@router.post(
    PRIVACY_REQUEST_RESUME_FROM_REQUIRES_INPUT,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestResponse,
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_CALLBACK_RESUME])
    ],
)
def resume_privacy_request_from_requires_input(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
) -> PrivacyRequestResponse:
    """Resume a privacy request from 'requires_input' status."""
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id
    )

    if privacy_request.status != PrivacyRequestStatus.requires_input:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Cannot resume privacy request from 'requires_input': privacy request '{privacy_request.id}' status = {privacy_request.status.value}.",
            # type: ignore
        )

    action_type = None
    if privacy_request.policy.get_rules_for_action(ActionType.access):
        action_type = ActionType.access
    elif privacy_request.policy.get_rules_for_action(ActionType.erasure):
        action_type = ActionType.erasure

    access_manual_webhooks: List[AccessManualWebhook] = AccessManualWebhook.get_enabled(
        db, action_type
    )
    try:
        for manual_webhook in access_manual_webhooks:
            # check the access or erasure cache based on the privacy request's action type
            if action_type == ActionType.access:
                privacy_request.get_manual_webhook_access_input_strict(manual_webhook)
            if action_type == ActionType.erasure:
                privacy_request.get_manual_webhook_erasure_input_strict(manual_webhook)
    except (
        NoCachedManualWebhookEntry,
        PydanticValidationError,
        ManualWebhookFieldsUnset,
    ) as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Cannot resume privacy request. {exc}",
        )

    logger.info(
        "Resuming privacy request '{}' after manual inputs verified",
        privacy_request_id,
    )

    privacy_request.status = PrivacyRequestStatus.in_processing
    privacy_request.save(db=db)
    queue_privacy_request(
        privacy_request_id=privacy_request.id,
    )

    return privacy_request  # type: ignore[return-value]


@router.get(
    REQUEST_TASKS,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ])],
    response_model=List[PrivacyRequestTaskSchema],
)
def get_individual_privacy_request_tasks(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
) -> List[RequestTask]:
    """Returns individual Privacy Request Tasks created by DSR 3.0 scheduler
    in order by creation and collection address"""
    pr: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id, error_if_deleted=False
    )

    logger.info(f"Getting Request Tasks for '{privacy_request_id}'")

    return pr.request_tasks.order_by(
        RequestTask.created_at.asc(), RequestTask.collection_address.asc()
    ).all()


@router.post(
    PRIVACY_REQUEST_REQUEUE,
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_CALLBACK_RESUME])
    ],
    response_model=PrivacyRequestResponse,
)
def requeue_privacy_request(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
) -> PrivacyRequestResponse:
    """
    Endpoint for manually re-queuing a stuck Privacy Request from selected states - use with caution.

    Don't use this unless the Privacy Request is stuck.
    """
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id
    )

    try:
        return _requeue_privacy_request(db, privacy_request)
    except PrivacyRequestError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.message,
        )


@router.post(
    REQUEST_TASK_CALLBACK,
    response_model=Dict,
)
def request_task_async_callback(
    *,
    data: RequestTaskCallbackRequest,
    request_task: RequestTask = Security(
        verify_request_task_callback, scopes=[PRIVACY_REQUEST_CALLBACK_RESUME]
    ),
) -> Dict:
    """
    For DSR 3.0: Async Callback Endpoint for Request Tasks

    Re-queues Request Task with asynchronously-retrieved results.  If you don't want to supply results, hitting
    this endpoint with an empty json request body will be taken as verification the async or erasure request
    for this request task is complete.
    """
    db = Session.object_session(
        request_task
    )  # Use the session created in the jwe verification

    privacy_request: PrivacyRequest = request_task.privacy_request
    if privacy_request.status not in [
        PrivacyRequestStatus.in_processing,
        PrivacyRequestStatus.approved,
    ]:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Callback failed. Cannot queue {request_task.action_type.value} task '{request_task.id}' with privacy request status '{privacy_request.status.value}'",
        )
    if request_task.status != ExecutionLogStatus.awaiting_processing:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Callback failed. Cannot queue {request_task.action_type.value} task '{request_task.id}' with request task status '{request_task.status.value}'",
        )
    logger.info(
        "Callback received for {} task {} {}",
        request_task.action_type.value,
        request_task.collection_address,
        request_task.id,
    )
    # Mark that the callback was received on the task itself.
    request_task.callback_succeeded = True

    if data.access_results and request_task.action_type == ActionType.access:
        # If access data should be added to results package, it should be
        # supplied in request body as a list of rows.  This data will be further filtered
        # by the policy before uploading to the end user
        request_task.access_data = data.access_results or []
    if data.rows_masked and request_task.action_type == ActionType.erasure:
        # For erasure requests, rows masked can be supplied here.
        request_task.rows_masked = data.rows_masked

    request_task.update_status(db, ExecutionLogStatus.pending)
    request_task.save(db)

    log_task_queued(request_task, "callback")
    queue_request_task(request_task, privacy_request_proceed=True)

    return {"task_queued": True}


@router.post(
    PRIVACY_REQUEST_BULK_SOFT_DELETE,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_DELETE])],
    status_code=HTTP_200_OK,
    response_model=BulkSoftDeletePrivacyRequests,
)
def bulk_soft_delete_privacy_requests(
    *,
    db: Session = Depends(deps.get_db),
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[PRIVACY_REQUEST_DELETE],
    ),
    privacy_requests: ReviewPrivacyRequestIds,
) -> BulkSoftDeletePrivacyRequests:
    """
    Soft delete a list of privacy requests. The requests' deleted_at field will be populated with the current datetime
    and its deleted_by field will be populated with the user_id of the user who initiated the deletion. Returns an
    object with the list of successfully deleted privacy requests and the list of failed deletions.
    """
    succeeded: List[str] = []
    failed: List[Dict[str, Any]] = []

    user_id = client.user_id
    if client.id == CONFIG.security.oauth_root_client_id:
        user_id = "root"

    for privacy_request_id in privacy_requests.request_ids:
        privacy_request = PrivacyRequest.get(db, object_id=privacy_request_id)

        if not privacy_request:
            failed.append(
                {
                    "message": f"No privacy request found with id '{privacy_request_id}'",
                    "data": {"privacy_request_id": privacy_request_id},
                }
            )
            continue

        if privacy_request.deleted_at is not None:
            failed.append(
                {
                    "message": f"Privacy request '{privacy_request_id}' has already been deleted.",
                    "data": {"privacy_request_id": privacy_request_id},
                }
            )
            continue

        privacy_request.soft_delete(db, user_id)
        succeeded.append(privacy_request.id)

    return BulkSoftDeletePrivacyRequests(succeeded=succeeded, failed=failed)


@router.post(
    PRIVACY_REQUEST_SOFT_DELETE,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_DELETE])],
    status_code=HTTP_200_OK,
    response_model=None,
)
def soft_delete_privacy_request(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[PRIVACY_REQUEST_DELETE],
    ),
) -> None:
    """
    Endpoint for soft deleting a privacy request. The request's deleted_at field will be populated with the current datetime
    and its deleted_by field will be populated with the user_id of the user who initiated the deletion.
    """
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id
    )
    user_id = client.user_id
    if client.id == CONFIG.security.oauth_root_client_id:
        user_id = "root"

    privacy_request.soft_delete(db, user_id)


@router.get(
    PRIVACY_REQUEST_ACCESS_RESULTS,
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ_ACCESS_RESULTS])
    ],
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestAccessResults,
)
def get_access_results_urls(
    privacy_request_id: str,
    *,
    db: Session = Depends(deps.get_db),
) -> PrivacyRequestAccessResults:
    """
    Endpoint for retrieving access results URLs for a privacy request.
    """
    if not CONFIG.security.subject_request_download_ui_enabled:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="Access results download is disabled.",
        )
    privacy_request: PrivacyRequest = get_privacy_request_or_error(
        db, privacy_request_id
    )

    if privacy_request.status != PrivacyRequestStatus.complete:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Access results for privacy request '{privacy_request_id}' are not available because the request is not complete.",
        )

    if not privacy_request.access_result_urls:
        return PrivacyRequestAccessResults(access_result_urls=[])

    return privacy_request.access_result_urls


@router.get(
    PRIVACY_REQUEST_FILTERED_RESULTS,
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ_ACCESS_RESULTS])
    ],
    status_code=HTTP_200_OK,
    response_model=FilteredPrivacyRequestResults,
)
def get_test_privacy_request_results(
    privacy_request_id: str,
    db: Session = Depends(deps.get_db),
) -> Dict[str, Any]:
    """Get filtered results for a test privacy request and update its status if complete."""
    privacy_request = get_privacy_request_or_error(db, privacy_request_id)

    if privacy_request.source != PrivacyRequestSource.dataset_test:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="Results can only be retrieved for test privacy requests.",
        )

    # Check completion status of all tasks
    dataset_key, statuses = get_task_info(privacy_request.access_tasks.all())
    all_completed = all(status in EXITED_EXECUTION_LOG_STATUSES for status in statuses)

    # Update request status if all tasks are done
    if all_completed:
        has_errors = ExecutionLogStatus.error in statuses
        privacy_request.status = (
            PrivacyRequestStatus.error if has_errors else PrivacyRequestStatus.complete
        )
        privacy_request.save(db=db)

    # Escape datetime and ObjectId values
    raw_data = privacy_request.get_raw_access_results()
    escaped_json = json.dumps(raw_data, indent=2, default=storage_json_encoder)
    results = json.loads(escaped_json)

    filtered_results: Dict[str, Any] = filter_access_results(
        db, results, dataset_key, privacy_request.policy_id  # type: ignore[arg-type]
    )

    with logger.contextualize(
        privacy_request_id=privacy_request.id,
        privacy_request_source=privacy_request.source.value,
    ):
        logger.info("Privacy request run completed.")

        return {
            "privacy_request_id": privacy_request.id,
            "status": privacy_request.status,
            "results": (
                filtered_results
                if CONFIG.security.dsr_testing_tools_enabled
                else "DSR testing tools are not enabled, results will not be shown."
            ),
        }


@router.post(
    PRIVACY_REQUEST_RESUBMIT,
    dependencies=[Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_CREATE])],
    response_model=PrivacyRequestResponse,
)
def resubmit_privacy_request(
    privacy_request_id: str,
    *,
    privacy_request_service: PrivacyRequestService = Depends(
        get_privacy_request_service
    ),
) -> PrivacyRequest:
    try:
        privacy_request = privacy_request_service.resubmit_privacy_request(
            privacy_request_id
        )
    except FidesopsException as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=exc.message
        )

    if not privacy_request:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail="Privacy request not found"
        )

    return privacy_request


def get_task_info(tasks: List[RequestTask]) -> Tuple[str, List[ExecutionLogStatus]]:
    """Returns first dataset and list of statuses from tasks"""
    statuses = []
    dataset_key = None

    for task in tasks:
        statuses.append(task.status)
        if dataset_key is None and task.dataset_name not in [
            "__ROOT__",
            "__TERMINATE__",
        ]:
            dataset_key = task.dataset_name

    return dataset_key, statuses  # type: ignore[return-value]


def filter_access_results(
    db: Session,
    access_results: Dict[str, Any],
    dataset_key: str,
    policy_id: str,
) -> Dict[str, List[Dict[str, Optional[Any]]]]:
    """Filter access results based on policy data categories.

    Args:
        access_results: Raw access results to filter
        dataset_key: Key of the dataset to get the data categories from
        policy_id: ID of the policy containing target data categories

    Returns:
        Filtered results containing only data matching policy target data categories
    """
    # Get dataset graph, we're assuming a 1-to-1 relationship between DatasetConfig and Dataset
    dataset_config = DatasetConfig.filter(
        db=db, conditions=(DatasetConfig.fides_key == dataset_key)
    ).first()
    if not dataset_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No dataset with fides_key '{dataset_key}'",
        )

    # Test privacy requests operate on one dataset at a time
    graph_dataset = dataset_config.get_graph()
    modified_graph_dataset = replace_references_with_identities(
        dataset_config.fides_key, graph_dataset
    )
    dataset_graph = DatasetGraph(modified_graph_dataset)

    # Get policy target categories
    policy = Policy.get_by(db, field="id", value=policy_id)
    if not policy:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No policy with ID '{policy_id}'",
        )

    target_categories: Set[str] = set()
    for rule in policy.get_rules_for_action(action_type=ActionType.access):
        for target in rule.targets:  # type: ignore[attr-defined]
            target_categories.add(target.data_category)

    return filter_data_categories(access_results, target_categories, dataset_graph)


@router.get(
    "/privacy-request/{privacy_request_id}/logs",
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_READ_ACCESS_RESULTS])
    ],
    status_code=HTTP_200_OK,
    response_model=List[LogEntry],
)
def get_test_privacy_request_logs(
    privacy_request_id: str,
    db: Session = Depends(deps.get_db),
) -> List[Dict[str, Any]]:
    """Get logs for a test privacy request."""
    privacy_request = get_privacy_request_or_error(db, privacy_request_id)

    if privacy_request.source != PrivacyRequestSource.dataset_test:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="Logs can only be retrieved for test privacy requests.",
        )

    if not CONFIG.security.dsr_testing_tools_enabled:
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="DSR testing tools are not enabled.",
        )

    # Get logs from Redis
    cache = get_cache()
    return cache.get_decoded_list(f"log_{privacy_request_id}") or []



"""add indexes to ctl_systems and privacydeclaration

Revision ID: 7f7c2b098f5d
Revises: 1af6950f4625
Create Date: 2023-11-21 18:52:34.508076

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7f7c2b098f5d"
down_revision = "1af6950f4625"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("CREATE EXTENSION IF NOT EXISTS pg_trgm;")
    op.create_index(
        "ix_ctl_systems_name",
        "ctl_systems",
        [sa.text("name gin_trgm_ops")],
        postgresql_using="gin",
    )


def downgrade():
    op.drop_index("ix_ctl_systems_name", table_name="ctl_systems")
    op.execute("DROP EXTENSION IF EXISTS pg_trgm;")
